#cython: profile=False

from .variant import Variant
from .utilities cimport split
from .utilities import split_cigar, get_local_reference, relative_aln_pos, most_common
from .localn import make_aligner, align, findall_indels, findall_mismatches, is_worth_realn

from .utilities cimport split


def find_by_normalization(
    target,
    pileup,
    window,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    basequalthresh=24,
    is_first_pass=True,
):
    """Search indels equivalent to the target indel

    Args:
        target (Variant): target indel
        pileup (list): a list of dictized read (dict)
    Returns:
        annoated pileup (list): a list of dictized read (dict)
    """

    pos = target.pos

    if target.is_indel:
        pileup = [is_target_by_normalization(read, target) for read in pileup]
        _pos = [read.get("observed_pos", pos) for read in pileup if read["is_target"]]
        pos = most_common(_pos) if _pos else pos
        are_read_ends = [read.get("is_read_end", False) for read in pileup if read["is_target"] and not read["is_dirty"]]
        if are_read_ends:
            read_end_evidence_only = all(are_read_ends)
        else:
            read_end_evidence_only = False
    else:
        is_single = (target.variant_type == "S")
        alt_bases = target.alt
        pileup = [is_substitute_target(read, pos, alt_bases, is_single) for read in pileup]
        read_end_evidence_only = False

    return target, pileup, gap_extension_penalty, pos, read_end_evidence_only

    ##target, extension_penalty_used = seek_larger_gapped_aln(
    #    target,
    #    pileup,
    #    window,
    #    match_score,
    #    mismatch_penalty,
    #    gap_open_penalty,
    #    gap_extension_penalty,
    #    basequalthresh,
    #    is_first_pass,
    #)

    #if is_first_pass and extension_penalty_used == 255:
    #    return find_by_normalization(
    #        target,
    #        pileup,
    #        window,
    #        match_score,
    #        mismatch_penalty,
    #        gap_open_penalty,
    #        gap_extension_penalty,
    #        basequalthresh=24,
    #        is_first_pass=False,
    #    )
    #else:
    #    return target, pileup, extension_penalty_used




def is_target_by_normalization(read, target):
    """Check if read contains an indel equivalent to target

    read (dict): dictized read
    target (Variant)
    """
    if read.get("is_target", False):
        return read
    else:
        read["is_target"] = False

    # trivial case
    if read["is_reference_seq"]:
        read["is_worth_realn"] = False
        read["mismatches"] = []
        return read

    for indel in read[target.variant_type]:
        if target == indel[-1]:

            pos = target.pos

            if avoid_left_aln(read, target):
                pos = indel[0] # pos observed in alignment
                read["observed_pos"] = pos

            read["is_target"] = True

            # trim clipped bases
            lt_offset = read["start_offset"]
            read["lt_flank"] = indel[1]
            read["lt_ref"] = indel[4]
            read["lt_qual"] = indel[6]

            read["indel_seq"] = indel[2]

            rt_offset = read["end_offset"]
            read["rt_flank"] = indel[3]
            read["rt_ref"] = indel[5]
            read["rt_qual"] = indel[7]

            read["lt_cigar"], read["rt_cigar"] = split_cigar(
                read["cigar_string"], pos, read["read_start"]
            )

            if min(len(read["lt_ref"]), len(read["rt_ref"])) / len(read["read_seq"]) < 0.15 or "H" in read["cigar_string"]:
                read["is_read_end"] = True
            else:
                read["is_read_end"] = False

    findall_mismatches(read)
    read["is_worth_realn"] = is_worth_realn(read, target, qual_lim=23)

    return read


def avoid_left_aln(read, target):
    #check if normalized pos should be used

    pos = target.pos #normalized
    if "N" in read["cigar_string"]:
        if  read["covering_subread"]:
            return pos < read["covering_subread"][0]
        else:
            return True

    return False


def is_substitute_target(read, target_pos, alt_bases, is_single):
    """this is experimental
    """
    if read.get("is_target", False):
        return read
    else:
        read["is_target"] = False

    if read["is_reference_seq"] or not read["is_covering"]:
        return read

    lt_end, rt_end = split(
                        read["read_seq"],
                        read["cigar_string"],
                        target_pos,
                        read["read_start"],
                        False,
                        False,
                     )

    lt_base = lt_end[-1] if len(lt_end) else ""
    if is_single:
        read["is_target"] = (lt_base == alt_bases)
    else:
        rt_fetch_len = min(len(alt_bases) - 1, len(rt_end))
        rt_bases = rt_end[: rt_fetch_len]
        read["is_target"] = match_mnv(alt_bases, lt_base, rt_bases)

    return read


def match_mnv(truth, lt_base, rt_bases):

    if lt_base and rt_bases:
        return truth == (lt_base + rt_bases)
    elif lt_base:
        return truth[0] == lt_base
    else:
        return truth[-len(rt_bases)] == rt_bases


def get_most_centered_read(target, pileup, target_annotated=True):

    most_centered_read = None
    center_score = 0

    if target_annotated:
        targetpileup = [
            read for read in pileup if read["is_target"] and not read["is_dirty"]
        ]
    else:
        targetpileup = [read for read in pileup if not read["is_dirty"]]

    if targetpileup:
        dist2center = [
            0.5
            - relative_aln_pos(
                read["ref_seq"], read["cigar_list"], read["aln_start"], target.pos,
            )
            for read in targetpileup
        ]

        abs_dist2center = [abs(i) for i in dist2center]
        most_central = min(abs_dist2center)
        most_centered_read = targetpileup[abs_dist2center.index(most_central)]
        center_score = dist2center[abs_dist2center.index(most_central)]

    return most_centered_read, center_score


def get_closest_gap(center_score, read_end, target, pileup):
    pos_look_up = {}
    read_look_up = {}
    for read in pileup:
        if (
            not read["is_reference_seq"]
            and read["is_covering"]
            and (read["D"] or read["I"])
        ):

            gaps = []
            if center_score >= 0:
                if (
                    read["aln_start"] < target.pos - len(read_end)
                    and read["is_covering"]
                ):
                    gaps = [
                        i[-1] for i in read["D"] + read["I"] if i[-1] != target
                    ]
            else:
                if read["aln_end"] > target.pos + len(read_end) and read["is_covering"]:
                    gaps = [
                        i[-1] for i in read["D"] + read["I"] if i[-1] != target
                    ]

            for i in gaps:
                if i in pos_look_up:
                    read_look_up[i].append(read)
                else:
                    pos_look_up[i] = abs(i.pos - target.pos)
                    read_look_up[i] = [read]
    if pos_look_up:
        closest_gap = min(pos_look_up, key=pos_look_up.get)
        closest_gap_reads = read_look_up[closest_gap]
        central_closest_gap_read, score = get_most_centered_read(
            closest_gap, closest_gap_reads, target_annotated=False,
        )

        if closest_gap and central_closest_gap_read:
            return closest_gap, central_closest_gap_read
        else:
            return None
    else:
        return None


def seek_larger_gapped_aln(
    target,
    pileup,
    window,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    basequalthresh,
    is_first_pass,
):

    read, center_score = get_most_centered_read(target, pileup)

    if not read:
        return target, gap_extension_penalty
    else:
        read_seq, ref_seq, cigarstring = (
            read["read"].query_alignment_sequence,
            read["ref_seq"],
            read["cigar_string"],
        )
        lt_read, rt_read = split(
            read_seq,
            cigarstring,
            target.pos,
            read["aln_start"],
            is_for_ref=True,
            reverse=False,
        )

        lt_ref, rt_ref = split(
            ref_seq,
            cigarstring,
            target.pos,
            read["aln_start"],
            is_for_ref=True,
            reverse=False,
        )

        lt_qual, rt_qual = split(
            read["read_qual"],
            cigarstring,
            target.pos,
            read["read_start"],
            is_for_ref=False,
            reverse=False,
        )

        if target.is_ins:
            rt_read = rt_read[len(target.indel_seq):]
        elif target.is_del:
            rt_ref = rt_ref[len(target.indel_seq):]

        with_end_mut = False
        if center_score >= 0:
            if lt_read != lt_ref and min(lt_qual) > basequalthresh:
                with_end_mut = True
        else:
            if rt_read != rt_ref and min(rt_qual) > basequalthresh:
                with_end_mut = True

    if is_first_pass and with_end_mut:

        read_end = lt_read if center_score >= 0 else rt_read

        if len(read_end) / len(read["read_seq"]) < 0.25:
            res = get_closest_gap(center_score, read_end, target, pileup)
            if res:
                closet_gap, closest_gap_read = res[0], res[1]
                subject_aligned_seq = closest_gap_read["read"].query_alignment_sequence
                query_aligned_seq = read["read"].query_alignment_sequence

                diff = len(query_aligned_seq) - len(subject_aligned_seq)
                if diff > 0:
                    if center_score >= 0:
                        query_aligned_seq = query_aligned_seq[:-diff]
                    else:
                        query_aligned_seq = query_aligned_seq[diff:]

                if read_end in query_aligned_seq and len(query_aligned_seq) > 30:
                    if query_aligned_seq in subject_aligned_seq:
                        return closet_gap, 255

    if "N" in read["cigar_string"]:
        ref_seq, lt_len = get_local_reference(target, [read], window)
    else:
        ref_seq, lt_len = get_local_reference(target, [read], window, unspliced=True)

    orig_gap_extension_penalty = gap_extension_penalty
    gap_extension_penalty = (
        0 if abs(center_score) > 0.35 and with_end_mut else gap_extension_penalty
    )
    aln = align(
        make_aligner(ref_seq, match_score, mismatch_penalty),
        read_seq,
        gap_open_penalty,
        gap_extension_penalty,
    )

    genome_aln_pos = target.pos + 1 - lt_len + aln.reference_start

    indels, mismatches = findall_indels(aln, genome_aln_pos, ref_seq, read_seq, report_snvs=True)
    if len(indels) != 1 or mismatches:
        return target, orig_gap_extension_penalty

    closest = min([abs(target.pos - indel["pos"]) for indel in indels])
    if "N" in read["cigar_string"] and closest > 3:
        return target, gap_extension_penalty

    candidates = [
        indel for indel in indels if abs(target.pos - indel["pos"]) == closest
    ]

    if candidates:
        candidate = candidates[0]
        if candidate["indel_type"] == "I":
            ref = candidate["lt_ref"][-1]
            alt = ref + candidate["indel_seq"]
        else:
            alt = candidate["lt_ref"][-1]
            ref = alt + candidate["del_seq"]

        target = Variant(target.chrom, candidate["pos"], ref, alt, target.reference)

    return target, gap_extension_penalty



# 786
# Aldy source: common.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Iterable, Any, List
import importlib_resources
import re
import time
import pprint
import logbook
import textwrap
import collections


PROTEINS = {
    "TTT": "F",
    "CTT": "L",
    "ATT": "I",
    "GTT": "V",
    "TTC": "F",
    "CTC": "L",
    "ATC": "I",
    "GTC": "V",
    "TTA": "L",
    "CTA": "L",
    "ATA": "I",
    "GTA": "V",
    "TTG": "L",
    "CTG": "L",
    "ATG": "M",
    "GTG": "V",
    "TCT": "S",
    "CCT": "P",
    "ACT": "T",
    "GCT": "A",
    "TCC": "S",
    "CCC": "P",
    "ACC": "T",
    "GCC": "A",
    "TCA": "S",
    "CCA": "P",
    "ACA": "T",
    "GCA": "A",
    "TCG": "S",
    "CCG": "P",
    "ACG": "T",
    "GCG": "A",
    "TAT": "Y",
    "CAT": "H",
    "AAT": "N",
    "GAT": "D",
    "TAC": "Y",
    "CAC": "H",
    "AAC": "N",
    "GAC": "D",
    "TAA": "X",
    "CAA": "Q",
    "AAA": "K",
    "GAA": "E",
    "TAG": "X",
    "CAG": "Q",
    "AAG": "K",
    "GAG": "E",
    "TGT": "C",
    "CGT": "R",
    "AGT": "S",
    "GGT": "G",
    "TGC": "C",
    "CGC": "R",
    "AGC": "S",
    "GGC": "G",
    "TGA": "X",
    "CGA": "R",
    "AGA": "R",
    "GGA": "G",
    "TGG": "W",
    "CGG": "R",
    "AGG": "R",
    "GGG": "G",
}
"""Codon table (stop codon is X)."""


REV_COMPLEMENT = {"A": "T", "T": "A", "C": "G", "G": "C"}
"""Reverse-complement DNA table."""


log = logbook.Logger("Aldy")
"""Default console logger."""


SOLUTION_PRECISION = 1e-2
"""
Solution precision (all values whose absolute difference falls below the specified
precision are considered equal).
"""


class AldyException(Exception):
    """Aldy exception class."""

    pass


class GRange(collections.namedtuple("GRange", ["chr", "start", "end"])):
    """Reference genome range (e.g. `chr22:10-20`). Immutable."""

    def samtools(self, pad_left=500, pad_right=1, prefix="") -> str:
        """Samtools-compatible region representation (e.g. chr1:10-20).

        :param pad_left: Left padding.
        :param pad_right: Right padding.
        :param prefix: Chromosome prefix."""

        return "{}:{}-{}".format(
            prefix + self.chr, max(1, self.start - pad_left), self.end + pad_right
        )

    def __str__(self):
        return self.samtools(0, 0, "")


def allele_name(x: str) -> str:
    """:returns: Major allele number of the star-allele name (e.g. `'12A'` -> `12`)."""
    if "*" in x:
        x = x.split("*", maxsplit=1)[1]
    return x.replace("/", "_")


def rev_comp(seq: str) -> str:
    """:returns: Reverse-complemented DNA sequence."""

    return "".join([REV_COMPLEMENT.get(x, x) for x in seq[::-1]])


def seq_to_amino(seq: str) -> str:
    """:returns: Protein sequence formed from the provided DNA sequence."""

    return "".join(
        PROTEINS[seq[i : i + 3]] for i in range(0, len(seq) - len(seq) % 3, 3)
    )


def freezekey(x):
    """Hashing support for dictionaries."""
    a = tuple(i[1] for i in sorted(x[0].items()))
    if len(x) > 1:
        a += tuple(i[1] for i in sorted(x[1].items()))
    return a


def sorted_tuple(x: Iterable) -> tuple:
    """:returns: Sorted tuple."""
    return tuple(sorted(x))


def td(s: str) -> str:
    """
    Abbreviation for textwrap.dedent. Used for stripping indentation in multi-line
    docstrings.
    """
    return textwrap.dedent(s)


class Timing:
    """
    Context manager for timing code blocks. Prints the time spent in the function after
    it is completed.
    """

    def __init__(self, name="Block", fn=None):
        self.name = name
        self.fn = fn if fn else log.debug

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *_):
        self.end = time.time()
        self.fn(f"{self.name} took {self.end - self.start:.2f}s")


def pp(x) -> str:
    """:returns: Pretty-printed variable string."""

    return pprint.pformat(x)


def script_path(key: str) -> str:
    """
    Obtain the full path of a resource.

    :param key: resource to be extracted.
    :param key: resource to be extracted in `path/file` format
        (e.g., `aldy.resources/test.txt`).
    :returns: Full path of the resource.
    :raises: :py:class:`aldy.common.AldyException` if the resource does not exist.
    """
    components = key.split("/")
    if len(components) < 2:
        raise AldyException(f'"{key}"" is not valid resource name')
    return str(importlib_resources.files(components[0]) / "/".join(components[1:]))


def colorize(text: str, color: str = "green") -> str:
    """:returns: xterm-compatible colorized string with a given color."""

    import logbook._termcolors

    return logbook._termcolors.colorize(color, text)


def parse_cn_region(cn_region):
    """
    :returns: :py:class:`GRange` object that represents the user-provided CN region in
        Samtools format (i.e., `chr1:100-200`).
    :raises: :py:class:`aldy.common.AldyException` if the region is invalid.
    """
    if cn_region is not None:
        r = re.match(r"^(.+?):(\d+)-(\d+)$", cn_region)
        if not r:
            raise AldyException(
                f"Parameter --cn-neutral={cn_region} cannot be parsed. "
                + "Must be chr:start-end (where start and end are numbers)"
            )
        ch = r.group(1)
        if ch.startswith("chr"):
            ch = ch[3:]
        return GRange(ch, int(r.group(2)), int(r.group(3)))
    return None


def chr_prefix(ch: str, chrs: List[str]) -> str:
    """
    Check if a chromosome needs "chr" prefix given the available chromosomes.
    :returns: Chromosome prefix if the chromosome does not have it.
    """
    if ch not in chrs and "chr" + ch in chrs:
        return "chr"
    return ""


class JsonDict(dict):
    """
    Dictionary that adds a dictionary for each missing key. Used to ease handling and
    populating JSON objects.
    """

    def __getitem__(self, key):
        if key not in self:
            self[key] = JsonDict()
        return self.get(key)


json: Any = JsonDict()



# 786
# Aldy source: diplotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Dict, Any
from natsort import natsorted
import collections
import re

from .common import td
from .gene import Gene, Mutation
from .coverage import Coverage
from .solutions import MinorSolution
from .version import __version__ as version


OUTPUT_COLS = [
    "Sample",
    "Gene",
    "SolutionID",
    "Major",
    "Minor",
    "Copy",
    "Allele",
    "Location",
    "Type",
    "Coverage",
    "Effect",
    "dbSNP",
    "Code",
    "Status",
]
"""Output column descriptions"""


def write_decomposition(
    sample: str, gene: Gene, coverage: Coverage, sol_id: int, minor: MinorSolution, f
):
    """Write an allelic decomposition to the given file.

    :param sample: Sample name.
    :param gene: Gene instance.
    :param sol_id: Solution ID (each solution must have a different ID).
    :param minor: Minor star-allele solution to be written.
    :param f: Output file.
    """

    for copy, a in enumerate(minor.solution):
        assert a.minor
        mutations = set(gene.alleles[a.major].func_muts) | set(
            gene.alleles[a.major].minors[a.minor].neutral_muts
        )
        mutations |= set(a.added)
        mutations -= set(a.missing)
        items = []
        if len(mutations) > 0:
            for m in sorted(mutations):
                fn = gene.get_functional(m, False)
                items.append(
                    [
                        sample,
                        gene.name,
                        sol_id,
                        minor.get_major_diplotype().replace(" ", ""),
                        ";".join(ay.minor for ay in minor.solution if ay.minor),
                        copy,
                        a.minor,
                        m.pos,
                        m.op,
                        coverage[m],
                        fn if fn else "none",
                        gene.get_rsid(m, default=False),
                        "",
                    ]
                )
        else:
            items.append(
                [
                    sample,
                    gene.name,
                    sol_id,
                    minor.get_major_diplotype().replace(" ", ""),
                    ";".join(ay.minor for ay in minor.solution if ay.minor),
                    copy,
                    a.minor,
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                ]
            )
        for it in items:
            print("\t".join(str(i) for i in it), file=f)


def write_vcf(
    sample: str, gene: Gene, coverage: Coverage, minors: List[MinorSolution], f
):
    """
    Write an allelic decomposition in the VCF format to the given file.

    :param sample: Sample name.
    :param gene: Gene instance.
    :param sol_id: Solution ID (each solution must have a different ID).
    :param minor: Minor star-allele solution to be written.
    :param f: Output file.
    """

    header = f"""
    ##fileformat=VCFv4.2
    ##source=aldy-v{version} (gene: {gene.name}-{gene.version})
    ##INFO=<ID=ANN,Number=1,Type=String,Description="Location within {gene.name}">
    ##INFO=<ID=TYPE,Number=1,Type=String,Description="Mutation kind">
    ##INFO=<ID=GENE,Number=1,Type=String,Description="Gene">
    ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
    ##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
    ##FORMAT=<ID=MA,Number=1,Type=String,Description="Major genotype star-allele calls">
    ##FORMAT=<ID=MI,Number=1,Type=String,Description="Minor genotype star-allele calls">
    """

    all_mutations: Dict[Mutation, List[dict]] = {
        m: [collections.defaultdict(int)] * len(minors)
        for minor in minors
        for a in minor.solution
        if a.minor
        for m in set(gene.alleles[a.major].func_muts)
        | set(gene.alleles[a.major].minors[a.minor].neutral_muts)
        | set(a.added)
    }
    for mi, minor in enumerate(minors):
        for ai, a in enumerate(minor.solution):
            assert a.minor
            mutations = set(gene.alleles[a.major].func_muts) | set(
                gene.alleles[a.major].minors[a.minor].neutral_muts
            )
            mutations |= set(a.added)
            for m in mutations:
                all_mutations[m][mi][ai] = 1
    print(
        td(header).strip()
        + "\n"
        + "\t".join(
            ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT"]
            + [
                f"{sample}:{mi}:{m.get_major_diplotype().replace(' ', '')}"
                for mi, m in enumerate(minors)
            ]
        ),
        file=f,
    )
    for m in sorted(all_mutations):
        ref = m.op[0]  # TODO: should be genome nucleotide, not the RefSeq nucleotide?!
        if m.op[1] == ">":
            alt = m.op[2]
        elif m.op[:3] == "ins":
            alt = ref + m.op[3:]
        else:
            ref, alt = ".", f"{m.op[3:]}, ."  # TODO: this is wrong?!

        fm = gene.get_functional(m)
        fm = fm.replace(" ", "_").replace("\t", "_").replace(";", "_") if fm else "none"
        info = [
            f"EFFECT={fm}",
            "GENE=" + gene.name,
        ]
        data = []
        for mi, minor in enumerate(minors):
            nall = len(minor.solution)
            data.append(
                {
                    "GT": "|".join(str(all_mutations[m][mi][i]) for i in range(nall)),
                    "DP": str(coverage[m]),
                    "MA": ",".join(
                        (
                            f"*{minor.solution[i].major}"
                            if all_mutations[m][mi][i] > 0
                            else "-"
                        )
                        for i in range(nall)
                    ),
                    "MI": ",".join(
                        (
                            f"*{minor.solution[i].minor}"
                            if all_mutations[m][mi][i] > 0
                            else "-"
                        )
                        for i in range(nall)
                    ),
                }
            )
        pattern = (
            "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t"
            + "{format}\t{data}"
        )
        print(
            pattern.format(
                chrom=gene.chr,
                pos=m.pos + 1,
                id=gene.get_rsid(m, default=False),
                ref=ref,
                alt=alt,
                qual=0,
                filter="PASS",
                info=";".join(info),
                format=":".join(data[0].keys()),
                data="\t".join(":".join(d.values()) for d in data),
            ),
            file=f,
        )


def estimate_diplotype(gene: Gene, solution: MinorSolution) -> str:
    """Calculate the diplotype assignment for a minor solution.
    Set the `diplotype` attribute of the :py:class:`aldy.solutions.MinorSolution`.

    Uses the diplotype assignment heuristics to assign correct diplotypes.
    This heuristics has no biological validity whatsoever---it is purely used
    for pretty-printing the final solutions.

    :returns: Diplotype assignment.
    """

    del_allele = gene.deletion_allele()

    # solution is the array of (major, minor) tuples
    major_dict: Dict[str, List[int]] = collections.defaultdict(list)
    for ai, a in enumerate(solution.solution):
        n = str(a.major).split("#")[0]  # chop off fusion suffix
        components = re.split(r"(\d+)", n)
        real = components[0] if components[0] != "" else components[1]
        major_dict[real].append(ai)
    if del_allele:
        if len(solution.solution) == 0:
            major_dict[del_allele].append(-1)
            major_dict[del_allele].append(-1)
        elif len(solution.solution) == 1:
            major_dict[del_allele].append(-1)

    diplotype: Any = [[], []]
    dc = 0

    # Handle tandems (heuristic that groups common tandems together,
    #                 e.g. 1, 2, 13 -> 13+1/2 if [13,1] is a common tandem)
    if len(solution.solution) > 2:
        for ta, tb in gene.common_tandems:
            while major_dict[ta] and major_dict[tb]:
                diplotype[dc % 2].append((major_dict[ta][0], major_dict[tb][0]))
                dc += 1
                del major_dict[ta][0], major_dict[tb][0]

    # Handle duplicates (heuristics that groups duplicate alleles together,
    #                    e.g. 1, 1, 2 -> 1+1/2)
    # First check should we split them (e.g. 1, 1, 1, 1 -> 1+1/1+1)?
    xlen = lambda d: sum(2 if isinstance(n, tuple) else 1 for n in d)  # noqa

    if len(major_dict) == 1:
        items = next(iter(major_dict.values()))
        if len(items) % 2 == 0:
            diplotype[dc % 2] += items[: len(items) // 2]
            dc += 1
            diplotype[dc % 2] += items[len(items) // 2 :]
            dc += 1
            major_dict.clear()
    for _, items in major_dict.items():
        if len(items) > 1:
            if xlen(diplotype[dc % 2]) > xlen(diplotype[(dc + 1) % 2]):
                dc += 1
            diplotype[dc % 2] += items
            items.clear()
            dc += 1

    # Handle the rest
    for _, items in major_dict.items():
        if items:
            if xlen(diplotype[dc % 2]) > xlen(diplotype[(dc + 1) % 2]):
                dc += 1
            assert len(items) == 1
            diplotype[dc % 2] += items
            dc += 1

    # Each diplotype should have at least one item
    # e.g. 1, 1 -> becomes 1+1/_ due to duplicate heuristic -> fixed to 1/1
    if len(diplotype[1]) == 0:
        if len(diplotype[0]) > 1:
            diplotype = diplotype[0][:-1], [diplotype[0][-1]]
        elif len(diplotype[0]) == 1 and isinstance(diplotype[0][0], tuple):
            diplotype = diplotype[0][0]

    def flatten(d):
        def key(x):
            if isinstance(x, tuple):
                return solution.get_major_name(x[0])
            return solution.get_major_name(x)

        for i in natsorted(d, key=key):
            if isinstance(i, tuple):
                yield from i
            else:
                yield i

    # Make sure that the elements are sorted and that the tandems are grouped together
    diplotype = natsorted(
        [list(flatten(diplotype[0])), list(flatten(diplotype[1]))],
        key=lambda x: [solution.get_major_name(y) for y in x],
    )
    solution.set_diplotype(diplotype)
    return diplotype


def estimate_cpic(gene: Gene, solution: MinorSolution) -> str:
    """Calculate the CPIC functionality for a minor solution.
    :returns: CPIC functionality.
    """

    activities = []
    for ai, a in enumerate(solution.solution):
        if a.extra_functionality():
            activities.append("unknown")
        else:
            activities.append(gene.alleles[a.major].minors[a.minor].activity)
    if "unknown" not in activities:
        if gene.cpic_scores:
            score = 0.0
            for a in activities:
                if a not in gene.cpic_scores:
                    break
                score += gene.cpic_scores[a]
            else:
                for fn, [lo, hi] in gene.cpic.items():
                    if lo <= score <= hi:
                        return (score, fn)
        else:
            activities = tuple(sorted(activities))
            i = [fn for fn, a in gene.cpic.items() if activities in a]
            if len(i) == 1:
                return (0, i[0])
    return (0, "indeterminate")



#cython: profile=False

import re
import numpy as np
from .sswpy import SSW

from .consensus import is_compatible
from .utilities import get_mapped_subreads, get_end_pos, make_insertion_first, to_minimal_repeat_unit

from .utilities cimport split, count_lowqual_non_ref_bases

cigar_ptrn = re.compile(r"[0-9]+[MIDNSHPX=]")


def find_by_smith_waterman_realn(
    target_indel,
    contig,
    pileup,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    basequalthresh,
    mapq_lim=1,
):
    """Annotate if reads contain target indel

    Args:
        target_indel (Variant)
        template (dict): indel template
        pileup (list): list of dictized reads (dict)
        mapq (int): minimum mapping quality
    Returns:
        pileup (list): annotated pileup
    """
    indel_type, indel_seq = target_indel.variant_type, target_indel.indel_seq

    mut_ref_lt, mut_ref_mid, mut_ref_rt = contig.get_contig_seq(split=True)
    ref_ref = contig.get_reference_seq()
    mut_ref = mut_ref_lt + mut_ref_mid + mut_ref_rt

    mut_aligner = make_aligner(mut_ref, match_score, mismatch_penalty)
    ref_aligner = make_aligner(ref_ref, match_score, mismatch_penalty)

    pileup = [findall_mismatches(read) for read in pileup]

    pileup = [
        is_target_by_ssw(
            read,
            target_indel,
            contig,
            mut_ref_lt,
            mut_ref_mid,
            mut_ref_rt,
            mut_aligner,
            ref_aligner,
            match_score,
            mismatch_penalty,
            gap_open_penalty,
            gap_extension_penalty,
            indel_type,
            basequalthresh,
            mapq_lim,
        )
        for read in pileup
    ]

    return pileup


def findall_mismatches(read, end_trim=0):

    if read["is_reference_seq"]:
        read["mismatches"] = []
        return read

    aln_start, aln_end = read["aln_start"], read["aln_end"]
    mapped_subpreads = get_mapped_subreads(read["cigar_string"], aln_start, aln_end)

    mismatches = []
    for subread in mapped_subpreads:
        start, end = subread[0], subread[1]
        span = end - start + 1

        # trim clipped segments
        cigarstring = read["cigar_string"]
        if "S" in cigarstring:
            cigarlst = read["cigar_list"]
            read_seq = read["read_seq"]
            quals = read["read_qual"]

            if "S" in cigarlst[0]:
                cigarlst = cigarlst[1:]
                read_seq = read_seq[read["start_offset"] :]
                quals = quals[read["start_offset"] :]

            if "S" in cigarlst[-1]:
                cigarlst = cigarlst[:-1]
                read_seq = read_seq[: -read["end_offset"]]
                quals = quals[: -read["end_offset"]]

            cigarstring = "".join(cigarlst)
        else:
            read_seq = read["read_seq"]
            quals = read["read_qual"]

        lt_seq, rt_seq = split(
            read_seq, cigarstring, start, aln_start, is_for_ref=False, reverse=False,
        )
        lt_qual, rt_qual = split(
            quals, cigarstring, start, aln_start, is_for_ref=False, reverse=False,
        )
        lt_ref, rt_ref = split(
            read["ref_seq"],
            cigarstring,
            start,
            aln_start,
            is_for_ref=True,
            reverse=False,
        )

        mapped_seq = lt_seq[-1] + rt_seq[: span - 1]
        mapped_qual = [lt_qual[-1]] + list(rt_qual[: span - 1])
        mapped_ref = lt_ref[-1] + rt_ref[: span - 1]

        pos = start  # pos for first elem of the rt side
        for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):
            if r != a:
                if aln_start + end_trim <= pos <= aln_end - end_trim:
                    mismatches.append((pos, r.upper(), a, q))

            pos += 1

    read["mismatches"] = mismatches

    return read


def is_worth_realn(read, target_indel, qual_lim=23):

    if read["covering_subread"]:
        is_covered = True
        covering_start, covering_end = (
            read["covering_subread"][0],
            read["covering_subread"][1],
        )
    else:
        is_covered = False
        if target_indel.is_ins:
            return False
        else:
            covering_start = target_indel.pos
            covering_end = covering_start + len(target_indel.ref)

    dist_to_left_end = target_indel.pos - read["aln_start"]
    dist_to_right_end = read["aln_end"] - target_indel.pos
    if dist_to_left_end < 0:
        is_lefty = True
    elif dist_to_right_end < 0:
        is_lefty = False
    else:
        is_lefty = (dist_to_left_end <= dist_to_right_end)

    start_cigar, end_cigar = read["cigar_list"][0], read["cigar_list"][-1]

    # start clipped
    if is_lefty and covering_start < read["aln_start"] <= covering_end and int(start_cigar[:-1]) > 2:
        return True

    # end clipped
    if not is_lefty and  covering_start <= read["aln_end"] < covering_end and int(end_cigar[:-1]) > 2:
        return True

    mismatches = [
        var
        for var in read["mismatches"]
        if covering_start <= var[0] <= covering_end and var[3] > qual_lim
    ]

    shiftable_pos = [v.pos for v in target_indel.generate_equivalents()]
    lt_pos, rt_pos = min(shiftable_pos), max(shiftable_pos)

    if lt_pos < rt_pos:
        if is_lefty:
            if lt_pos < read["aln_start"]:
                lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
                lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
                if lt_end_read == lt_end_ref:
                    return False
        else:
            if read["aln_end"] <= rt_pos:
                rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]
                rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
                if rt_end_read == rt_end_ref:
                    return False


    if mismatches:
        if is_lefty:
            lt_most_pos = min(var[0] for var in mismatches)
            with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
        else:
            rt_most_pos = max(var[0] for var in mismatches)
            with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4

        if with_end_mismatches:
            return True
        else:
            if is_covered:
                return True
            else:
                return False

    indels = [
        var for var in read["I"] + read["D"] if covering_start <= var[0] <= covering_end
    ]

    if indels:
        return True

    return False

def is_target_by_ssw(
    read,
    target_indel,
    contig,
    mut_ref_lt,
    mut_ref_mid,
    mut_ref_rt,
    mut_aligner,
    ref_aligner,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    indel_type,
    basequalthresh,
    mapq_lim,
    mapped_base_cnt_thresh=40,
    allow_mismatches=10,
):

    # already found
    if read["is_target"]:
        return read

    if read["is_reference_seq"] or read["mapq"] <= mapq_lim or not is_worth_realn(read, target_indel):
        read["is_target"] = False
        return read

    read_seq = read["read_seq"]

    ref_aln = align(ref_aligner, read_seq, gap_open_penalty, gap_extension_penalty)
    # forced to align by setting gap_open_penalty=len(read_seq)
    mut_aln = align(mut_aligner, read_seq, len(read_seq), gap_extension_penalty)

    if mut_aln.optimal_score <= ref_aln.optimal_score:
        read["is_target"] = False
        return read

    mut_aln_cigar = mut_aln.CIGAR
    mut_aln_ref_start = mut_aln.reference_start
    mut_aln_ref_end = mut_aln.reference_end
    mut_aln_read_start = mut_aln.read_start
    mut_aln_read_end = mut_aln.read_end
    mut_cigar_list = cigar_ptrn.findall(mut_aln.CIGAR)

    target_covered = is_covering_target(
                    read["read_name"],
                    read_seq,
                    target_indel.indel_seq,
                    mut_ref_lt,
                    mut_ref_mid,
                    mut_ref_rt,
                    mut_aln_cigar,
                    len(read_seq),
                    mut_aln_ref_start,
                    mut_aln_ref_end,
                    mut_aln_read_start,
                    mut_aln_read_end,
                    target_indel.count_repeats(),
                   )

    if target_covered == 1:
        read["is_target"] = True
        return read
    elif target_covered == -1:
        read["undetermined"] = True
        return read
    else:
        return read

cdef int is_covering_target(
    str readname,
    str read_seq,
    str indel_seq,
    str mut_ref_lt,
    str mut_ref_mid,
    str mut_ref_rt,
    str mut_aln_cigar,
    int read_seq_len,
    int ref_aln_start,
    int ref_aln_end,
    int read_aln_start,
    int read_aln_end,
    int n_repeats,
):
    cdef list mut_cigar_list = cigar_ptrn.findall(mut_aln_cigar)

    if len(mut_cigar_list) > 1:
        return False

    cdef str repeat_unit = to_minimal_repeat_unit(indel_seq)
    cdef int mapped_len = int(mut_cigar_list[0][:-1])
    cdef int mut_ref_lt_len = len(mut_ref_lt)
    cdef int mut_ref_mid_len = len(mut_ref_mid)
    cdef int mut_ref_rt_len = len(mut_ref_rt)
    cdef int lt_consumption
    cdef int rt_consumption
    cdef int total_consumption = read_aln_end - read_aln_start + 1
    cdef bint is_lt_indel_end_reached
    cdef bint is_rt_indel_end_reached
    cdef bint is_lt_read_consumed = (read_aln_start == 0)
    cdef bint is_rt_read_consumed = (read_aln_end == read_seq_len - 1)

    # not covering indel
    if ref_aln_end < mut_ref_lt_len:
        return False
    elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:
        return False

    # ins
    if mut_ref_mid_len:
        if ref_aln_start < mut_ref_lt_len:
            lt_consumption = mut_ref_lt_len - ref_aln_start

            if total_consumption > lt_consumption + mut_ref_mid_len:
                lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
                mid_read = read_seq[read_aln_start + lt_consumption : read_aln_start + lt_consumption + mut_ref_mid_len]
                rt_read = read_seq[read_aln_start + lt_consumption + mut_ref_mid_len : read_aln_end + 1]

                lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)
                rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)

                if lt_check and rt_check:
                    #True
                    return 1
                else:
                    #False (undetermined)
                    return -1
            else:
                ### abolish heuristic for reapeat
                #if n_repeats:
                #    return False
                if is_rt_read_consumed:
                        #True
                        return 1
                else:
                    # aligned from left and ended in inserted seq
                    rt_consumption = total_consumption - lt_consumption
                    consumed_mid_seq = mut_ref_mid[:rt_consumption:]

                    if consumed_mid_seq == read_seq[-rt_consumption:]:
                        #True
                        return 1
                    else:
                        #False
                        return 0
        else:
            # no left side alignment

            ### abolish heuristic for repeat
            # instead check the exact match
            #if n_repeats:
            #    return False

            if is_lt_read_consumed:
                #True
                return 1
            else:
                # aligned from right and ended in inserted seq
                lt_consumption = mut_ref_lt_len + mut_ref_mid_len - ref_aln_start
                consumed_mid_seq = mut_ref_mid[-lt_consumption:]

                # require exact match
                if consumed_mid_seq == read_seq[:lt_consumption]:
                    #True
                    return 1
                else:
                    #False
                    return 0

    else:
        lt_consumption = mut_ref_lt_len - ref_aln_start
        rt_consumption = total_consumption - lt_consumption

        lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
        rt_read = read_seq[read_aln_start + lt_consumption  : read_aln_end]

        lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)
        rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)

        if lt_check and rt_check:
            pass
        else:
            #False (undetermined)
            return -1

        if lt_consumption <= rt_consumption:
            if is_lt_read_consumed:
                #True
                return 1
            else:
                if lt_consumption > 2:
                    #True
                    return 1
                else:
                    #False
                    return 0
        else:
            if is_rt_read_consumed:
                #True
                return 1
            else:
                if rt_consumption > 2:
                    #True
                    return 1
                else:
                    #False
                    return 0


def is_compatible_repeats(seq, repeat_unit, expected_n_repeats, is_left):
    unit_len = len(repeat_unit)


    if is_left:
        seq = seq[::-1]
        repeat_unit = repeat_unit[::-1]

    cnt = 0
    is_done = False
    while not is_done:
        if not seq:
            is_done = True
        elif repeat_unit == seq[: unit_len]:
            seq = seq[unit_len :]
            cnt += 1
        else:
            is_done = True

    is_compatible = True
    if not seq:
        is_compatible = False
    else:
        if cnt and cnt != expected_n_repeats:
            is_compatible = False

    return is_compatible




def make_aligner(ref_seq, match_score, mismatch_penalty):
    aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
    aligner.setReference(ref_seq)
    return aligner


def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):
    aligner.setRead(read_seq)
    return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)


def parse_read_by_mut_aln(mut_aln, contig, read, indel_type):
    """Decompose read based on the SSW alignment result for evaluation

    Args:
        mut_aln (named tuple): Alignment(CIGAR, optimal_score, sub_optimal_score, reference_start, reference_end, read_start, read_end)
        template (dict): indel template
        read (dict): dictized read
        indel_type (str): "I" for insertion or "D" for deletion
    Returns:
        read (dict): dictized read with read seq decomposed
    """
    lt_len, indel_len, rt_len = (
        len(contig.lt_consensus_seq),
        len(contig.indel_seq),
        len(contig.rt_consensus_seq),
    )

    read_seq = read["read_seq"]
    read_qual = read["read_qual"]
    ref_start, ref_end = mut_aln.reference_start, mut_aln.reference_end
    aln_start, aln_end = mut_aln.read_start, mut_aln.read_end

    lt_flank, mid_seq, rt_flank = "", "", ""
    lt_qual, rt_qual = [], []

    if ref_start <= lt_len:
        lt_diff = lt_len - ref_start
        lt_flank = read_seq[aln_start : aln_start + lt_diff]
        lt_qual = read_qual[aln_start : aln_start + lt_diff]
        if indel_type == "I":
            end_point = min(aln_start + lt_diff + indel_len, aln_end)
            mid_seq = read_seq[aln_start + lt_diff : end_point]
        else:
            rt_flank = read_seq[aln_start + lt_diff :]
            rt_qual = read_qual[aln_start + lt_diff :]
            del_pos = get_end_pos(
                read["read_start"] + aln_start, lt_flank, read["cigar_string"]
            )

            lt_ref, rt_ref = split(
                read["ref_seq"],
                read["cigar_string"],
                del_pos,
                read["aln_start"],
                is_for_ref=True,
                reverse=False,
            )

            read["del_pos"] = del_pos
            read["del_seq"] = rt_ref[:indel_len]

    if lt_len + indel_len <= ref_end and indel_type == "I":
        rt_diff = ref_end - (lt_len + indel_len)
        rt_flank = read_seq[aln_end - rt_diff : aln_end]
        rt_qual = read_qual[aln_end - rt_diff : aln_end]
        end_point = max(aln_start, aln_end - rt_diff - indel_len)
        mid_seq = read_seq[end_point : aln_end - rt_diff]

    read["lt_flank"] = lt_flank
    read["lt_qual"] = lt_qual
    read["indel_seq"] = mid_seq
    read["rt_flank"] = rt_flank
    read["rt_qual"] = rt_qual

    return read


def findall_indels(ref_aln, genome_aln_pos, ref_seq, read_seq, report_snvs=False, basequals=None):

    genome_aln_pos -= 1
    ref_idx = ref_aln.reference_start
    read_idx = ref_aln.read_start

    lt_clipped = read_seq[:read_idx]

    indels, snvs = [], []

    for token in cigar_ptrn.findall(make_insertion_first(ref_aln.CIGAR)):
        event, event_len = token[-1], int(token[:-1])

        if event == "I" or event == "D":
            indel = {}

            indel["pos"] = genome_aln_pos
            indel["lt_ref"] = ref_seq[:ref_idx]
            indel["lt_flank"] = read_seq[:read_idx]

            if basequals:
                indel["lt_qual"] = basequals[:read_idx]

            if event == "I":
                indel["indel_type"] = "I"
                indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]
                indel["rt_ref"] = ref_seq[ref_idx:]
                indel["rt_flank"] = read_seq[read_idx + event_len :]
                indel["ref_idx"] = ref_idx
                indel["read_idx"] = read_idx

                if basequals:
                    indel["rt_qual"] = basequals[read_idx + event_len :]

                read_idx += event_len
            else:
                indel["indel_type"] = "D"
                indel["indel_seq"] = ""
                indel["del_seq"] = ref_seq[ref_idx : ref_idx + event_len]
                indel["rt_ref"] = ref_seq[ref_idx + event_len :]
                indel["rt_flank"] = read_seq[read_idx:]
                indel["ref_idx"] = ref_idx
                indel["read_idx"] = read_idx

                if basequals:
                    indel["rt_qual"] = basequals[read_idx:]

                ref_idx += event_len
                genome_aln_pos += event_len

            indels.append(indel)

        else:
            if report_snvs:
                i = 0
                while i < event_len:
                     if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:
                         snv = {}
                         snv["pos"] = genome_aln_pos + i + 1
                         snv["ref"] = ref_seq[ref_idx + i : ref_idx + i + 1]
                         snv["alt"] = read_seq[read_idx + i : read_idx + i + 1]

                         snvs.append(snv)

                     i += 1

            ref_idx += event_len
            read_idx += event_len
            genome_aln_pos += event_len


    rt_clipped = read_seq[read_idx:]
    for indel in indels:
        indel["lt_clipped"] = lt_clipped
        indel["rt_clipped"] = rt_clipped

    if report_snvs:
        return indels, snvs
    else:
        return indels



# 786
# Aldy source: profile.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, List, Tuple, Optional
import pysam
import os
import os.path
import yaml

from collections import defaultdict
from natsort import natsorted
from .common import log, GRange, AldyException, script_path, chr_prefix
from .gene import Gene


class Profile:
    """Profile and model parameter information."""

    def __init__(self, name, cn_region=None, data=None, **kwargs):
        self.name = name
        """Name of the profile."""

        self.cn_region = cn_region
        """Location of the copy-number neutral region."""

        self.data = data
        """Profile coverage data."""

        self.gap = 0.0
        """
        Optimality gap. Non-zero values enable non-optimal solutions.
        Default: 0 (only optimal solutions)
        """

        self.cn_solution = None
        """
        User-specified copy-number configuration.
        Default: `None` (uses CN solver in :py:mod:`aldy.cn` for detection)
        """

        self.neutral_value = 0.0
        """
        Joint coverage of the copy-number neutral region.
        Default: 0 (typically specified in the profile's YAML file)
        """

        self.threshold = 0.5
        """
        Single-copy variant threshold.
        Its value indicate the fraction of total reads that contain the given variant
        in a single gene copy. For example, if two copies are given (maternal and
        paternal), and if the single copy coverage is 10, threshold of 0.5 will ensure
        that all variants with coverage less than 5 (i.e., 0.5 * 10) are filtered out.
        Default: 0.5
        """

        self.min_coverage = 2.0
        """
        Minimum coverage needed to call a variant.
        Default: 2 (5 for illumina/wgs)
        """

        self.min_quality = 10
        """
        Minimum base quality for a read base to be considered.
        Default: 10
        """

        self.min_mapq = 10
        """
        Minimum mapping quality for a read to be considered.
        Default: 10
        """

        self.phase = True
        """
        Set if the phasing model in :py:mod:`aldy.minor` is to be used.
        Default: `True`
        """

        self.sam_long_reads = False
        """
        Set if long reads should be split-mapped. Should be used when dealing
        with long PacBio or Nanopore reads
        Default: `False` (typically specified in the profile's YAML file)
        """

        self.sam_mappy_preset = "map-hifi"
        """
        Mappy preset to use for split-mapping.
        Default: map-hifi
        """

        self.cn_max = 20
        """
        Maximum possible copy number of a gene.
        Default: 20
        """

        self.cn_pce_penalty = 2.0
        """
        Error penalty applied to the PCE region during CN calling (1 for no penalty).
        Default: 2.0
        """

        self.cn_diff = 10.0
        """
        The first CN objective term (coverage fit) coefficient.
        Default: 10.0
        """

        self.cn_fit = 1.0
        """
        The second CN objective term (gene fit) coefficient.
        Default: 1.0
        """

        self.cn_parsimony = 0.5
        """
        The third CN objective term (max. parsimony) coefficient.
        Default: 0.5
        """

        self.cn_fusion_left = 0.5
        """
        Extra penalty for the left fusions.
        Default: 0.5.
        """

        self.cn_fusion_right = 0.25
        """
        Extra penalty for the right fusions.
        Default: 0.25.
        """

        self.major_novel = 21.0
        """
        Penalty for novel functional mutation (0 for no penalty).
        Should be large enough to avoid calling novel mutations unless really necessary.
        Default: 21.0 (i.e., `max_cn + 1`)
        """

        self.minor_miss = 1.5
        """
        Penalty for missed minor mutations (0 for no penalty).
        Ideally larger than `minor_add` as additions should be cheaper.
        Default: 1.5
        """

        self.minor_add = 1.0
        """
        Penalty for novel minor mutations (0 for no penalty).
        Zero penalty ensures that extra mutations are preferred over the coverage errors
        if the normalized variant slack coverage is >= 50%. Penalty of 1.0 prefers
        additions only if the variant slack coverage is >= 75%.
        Default: 1.0
        """

        self.minor_phase = 0.4
        """
        The minor star-allele calling model's phasing term coefficient.
        Default: 0.4
        """

        self.minor_phase_vars = 3_000
        """
        Number of variables to use during the phasing. Use lower number if the model
        takes too long to complete.
        Default: 3,000
        """

        self.male = False
        """
        Set if the sample is male (i.e., has two X chromosomes). Used for calling
        sex chromosome genes (e.g., G6PD) when the CN calling is disabled.
        Default: False
        """

        self.max_minor_solutions = 1
        """
        Maximum number of minor solutions to report for each major solution.
        Default: 1
        """

        self.display_format = False
        """
        New novel allele display format.
        Default: False
        """

        self.debug_probe = ""
        """
        (Debug) Show raw data for a given mutation (e.g., I223M)
        """

        self.debug_novel = False
        """
        (Debug) Show potential novel functional mutations that are not in the database.
        """

        self.min_avg_coverage = 2.0
        """
        Minimum average gene coverage needed for Aldy.
        Default: 2
        """

        self.vcf_sample_idx = 0
        """
        VCF sample index.
        Default: 0
        """

        self.indelpost = True
        """
        Use indelpost for indel realignment (unless sam_long_reads is set).
        Default: True
        """

        self.update(kwargs)

    def update(self, kwargs):
        params = {}
        for n, v in kwargs.items():
            if v is not None and n in self.__dict__:
                if n == "cn_solution":
                    self.__dict__[n] = v
                else:
                    try:
                        if isinstance(self.__dict__[n], bool):
                            self.__dict__[n] = not (v in ["False", "0"])
                        else:
                            typ = type(self.__dict__[n])
                            self.__dict__[n] = typ(v)
                    except (ValueError, TypeError):
                        raise AldyException(f"Invalid parameter {n}: {v}") from None
                params[n] = self.__dict__[n]
        ps = "; ".join(f"{n}={v}" for n, v in params.items())
        log.debug(f"[params] {ps}")
        return params

    @staticmethod
    def load(gene, profile, cn_region=None, **params):
        """
        Load the copy number profile and parameters from a profile file.

        :param gene: Gene instance.
        :param profile: A profile YAML or a SAM/BAM/CRAM file that contains profile
            data.
        :param cn_region: Copy-number neutral region.
        """
        prof = None
        is_yml = True
        if os.path.exists(profile) and os.path.isfile(profile):
            ext = os.path.splitext(profile)
            if ext[-1] in [".bam", ".sam", ".cram"]:
                assert gene and cn_region
                regions = {
                    (gene.name, r, gi): rng
                    for gi, gr in enumerate(gene.regions)
                    for r, rng in gr.items()
                }
                prof = Profile.get_sam_profile_data(
                    profile,
                    regions=regions,
                    genome=gene.genome,
                    cn_region=cn_region,
                )
                is_yml = False
            else:
                with open(profile) as f:
                    prof = yaml.safe_load(f)
        else:
            profile_path = script_path(
                "aldy.resources.profiles/{}.yml".format(profile.lower())
            )
            with open(profile_path) as f:
                prof = yaml.safe_load(f)
        if not prof:
            raise AldyException(f"Could not load profile from {profile}")
        if is_yml and profile != "illumina" and cn_region:
            raise AldyException("-n is only valid with illumina or BAM profile")

        if is_yml and profile == "illumina" and cn_region:
            prof["neutral"]["value"] = cn_region.end - cn_region.start
        if "neutral" not in prof or "value" not in prof["neutral"]:
            raise AldyException("Profile missing neutral region")
        if gene.name not in prof:
            raise AldyException(f"Profile missing {gene.name}")
        if gene.genome not in prof["neutral"]:
            raise AldyException(f"Profile {profile} not compatible with {gene.genome}")
        if cn_region is None:
            cn_region = GRange(*prof["neutral"][gene.genome])
        return Profile(
            profile,
            cn_region,
            prof,
            neutral_value=prof["neutral"].get("value"),
            **dict(prof.get("options", {}), **params),
        )

    @staticmethod
    def get_sam_profile_data(
        sam_path: str,
        ref_path: Optional[str] = None,
        regions: Dict[Tuple[str, str, int], GRange] = dict(),
        cn_region: Optional[GRange] = None,
        genome: Optional[str] = "hg19",
        params: Dict = dict(),
    ) -> Dict[str, Dict[str, List[float]]]:
        """
        Load the profile information from a SAM/BAM/CRAM file.

        :param regions: List of regions to be extracted.
        :param cn_region: Copy-number neutral region.

        :return: list of tuples `(gene_name, chromosome, loci, coverage)`.


        .. note::
            Profile samples used in the original Aldy paper:

                1. PGRNseq-v1/v3: NA17642
                2. PGRNseq-v2: NA19789
                3. Illumina: by definition contains all ones (uniform coverage).
        """

        if not genome:
            genome = "hg19"
        if len(regions) == 0:
            import importlib_resources

            gene_regions = {}
            for g in sorted(
                importlib_resources.files("aldy.resources.genes").iterdir()
            ):
                if g.suffix != ".yml":
                    continue
                log.debug("Loading {}...", g)
                try:
                    gg = Gene(script_path(f"aldy.resources.genes/{g}"), genome=genome)
                    for gi, gr in enumerate(gg.regions):
                        for r, rng in gr.items():
                            gene_regions[gg.name, r, gi] = rng
                except AldyException as e:
                    log.warn(e)

        else:
            gene_regions = regions

        default_cn_neutral_region = {
            "hg19": GRange("22", 42547463, 42548249),
            "hg38": GRange("22", 42151472, 42152258),
        }
        gene_regions["neutral", "value", 0] = (
            cn_region if cn_region else default_cn_neutral_region[genome]
        )

        chr_regions: Dict[str, Tuple[int, int]] = {}
        for c, s, e in gene_regions.values():
            if c not in chr_regions:
                chr_regions[c] = (s, e)
            else:
                chr_regions[c] = (min(s, chr_regions[c][0]), max(e, chr_regions[c][1]))

        cov: dict = defaultdict(lambda: defaultdict(int))
        for c, (s, e) in natsorted(chr_regions.items()):
            if sam_path == "<illumina>":
                continue
            with pysam.AlignmentFile(  # type: ignore
                sam_path, reference_filename=ref_path
            ) as sam:
                region = GRange(c, s, e).samtools(
                    pad_left=1000,
                    pad_right=1000,
                    prefix=chr_prefix(c, [x["SN"] for x in sam.header["SQ"]]),
                )
                log.info("Scanning {}...", region)
                try:
                    for read in sam.fetch(region=region):
                        start, s_start = read.reference_start, 0
                        if not read.cigartuples:
                            continue

                        for op, size in read.cigartuples:
                            if op == 2:
                                for i in range(size):
                                    cov[c][start + i] += 1
                                start += size
                            elif op == 1:
                                s_start += size
                            elif op == 4:
                                s_start += size
                            elif op in [0, 7, 8]:
                                for i in range(size):
                                    cov[c][start + i] += 1
                                start += size
                                s_start += size
                except ValueError:
                    log.warn("Cannot fetch {}", region)

        d: Dict = {}
        for (g, r, ri), (c, s, e) in gene_regions.items():
            if g not in d:
                d[g] = {}
            if r not in d[g]:
                d[g][r] = [0]
            if ri >= len(d[g][r]):
                d[g][r].append(0)
            if sam_path == "<illumina>":
                d[g][r][ri] = e - s
            else:
                d[g][r][ri] = sum(cov[c][i] for i in range(s, e))
        d["neutral"]["value"] = d["neutral"]["value"][0]
        d["neutral"][genome] = [*gene_regions["neutral", "value", 0]]
        if params:
            d["options"] = {}
            for k, v in Profile("").update(params).items():
                d["options"][k] = v
        return d



#cython: embedsignature=True
#cython: profile=False

cimport cython
import random
import numpy as np
from functools import partial

from .pileup import (
    update_pileup,
    retarget,
    update_read_info,
    check_overhangs,
    filter_spurious_overhangs,
)
from .gappedaln import find_by_normalization
from .softclip import find_by_softclip_split
from .localn import find_by_smith_waterman_realn, make_aligner

from .alleles import phase_nearby_variants
from .contig import compare_contigs
from .utilities import (
    get_local_reference,
    relative_aln_pos,
    split_cigar,
    most_common_gap_ptrn,
    get_gap_ptrn2,
    most_common
)
from .pileup cimport make_pileup
from .utilities cimport split
from .contig cimport Contig, FailedContig

from .variant cimport Variant, NullVariant
from .local_reference cimport UnsplicedLocalReference

from pysam.libcalignmentfile cimport AlignmentFile

random.seed(123)

cdef class VariantAlignment:
    """This class accepts the target indel as :class:`~indelpost.Variant`
    and the BAM file as `pysam.AlignmentFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignmentFile>`__ .
    Upon creation, the target indel is searched through realignment.
    Evaluation of the equality between :class:`~indelpost.VariantAlignment` objects triggers local-phasing around the indels
    to test the alignment identity.


    Parameters
    ----------
    variant : Variant
        :class:`~indelpost.Variant` object representing the target indel.

    bam : pysam.AlignmentFile
        BAM file supplied as
        `pysam.AlignmentFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignmentFile>`__ object.

    window : integer
        analyzes region input_indel_pos +/- window (defalt 50).

    exclude_duplicates : bool
        True (default) to exclude reads marked duplicate by
        `MarkDuplicate <https://broadinstitute.github.io/picard/command-line-overview.html#MarkDuplicates>`__.

    downsample_threshold : integer
        downsamples to the threshold if the covearge at the input locus exceeds the threshold (default to 1000).

    mapping_quality_threshold : integer
        reads with a mapping quality (MAPQ) < the threshold (default 1) will be filtered.

    base_quality_threshold : integer
        non-reference base-calls with a quality score < the threshold (default 20) will be masked as "N" in the analysis.

    retarget_search_window : integer
        attempts to re-target in input_indel_pos +/- window (defalt 30) if the exact match is not found after realignment.

    retarget_similarity_cutoff : float
        re-targets to the indel sequence with a highest
        `Ratcliff/Obershelp similarity score <https://docs.python.org/3/library/difflib.html#difflib.get_close_matches>`__ > cutoff (default 0.7).

    exact_match_for_shiftable : bool
        True to require exact (equivalent)  match for indels that are equivalently alignable by shifting the position (default True)

    match_score : integer
        score (default 3) for matched bases in Smith-Waterman local alignment.

    mismatch_penalty : interger
        penalty (default 2) for mismatched bases in Smith-Waterman local alignment.

    gap_open_penalty : integer
        penalty (default 3) to create a gap in Smith-Waterman local alignment.

    gap_extension_penalty : integer
        penalty (default 1) to expend a created gap by one base in Smith-Waterman local alignment.

    auto_adjust_extension_penalty : bool
        True (default) to auto-adjust the gap open and extend penalties to find the input indel by realignment.

    no_realignment : bool
        True to only analyzed gap-aligned indels (default False)
    """
    def __cinit__(
        self,
        Variant target,
        AlignmentFile bam,
        int window=50,
        bint exclude_duplicates=True,
        int retarget_search_window=30,
        float retarget_similarity_cutoff=0.7,
        bint exact_match_for_shiftable=True,
        int mapping_quality_threshold=1,
        int downsample_threshold=1000,
        int base_quality_threshold=20,
        int match_score=3,
        int mismatch_penalty=2,
        int gap_open_penalty=3,
        int gap_extension_penalty=1,
        bint auto_adjust_extension_penalty=True,
        bint no_realignment=False,
    ):

        self.target, second_target = target, target

        is_complex_input = False
        if not target.is_non_complex_indel() and target.is_indel:
            is_complex_input = True

            if auto_adjust_extension_penalty:
                decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty)
            else:
                decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty, gap_open_penalty, gap_extension_penalty)

            decomposed_indels = [i for i in decomposed_variants if i.is_indel]
            decomposed_indels.sort(key=lambda x : len(x.indel_seq))
            self.__target = decomposed_indels[-1]
            self.target = self.__target
            if len(decomposed_indels) > 1:
                second_target = decomposed_indels[-2]

            #self.__target = max(decomposed_indels, key=lambda l : len(l.indel_seq))
            #self.target = self.__target
        else:
            self.__target = target.normalize()

        self.bam = bam
        self.window = window
        self.exclude_duplicates = exclude_duplicates
        self.retarget_window = retarget_search_window
        self.retarget_cutoff = retarget_similarity_cutoff
        self.exact_match_for_shiftable = exact_match_for_shiftable
        self.mapqthresh = mapping_quality_threshold
        self.downsamplethresh = downsample_threshold
        self.basequalthresh = base_quality_threshold
        self.match_score = match_score
        self.mismatch_penalty = mismatch_penalty
        self.gap_open_penalty = gap_open_penalty
        self.gap_extension_penalty = gap_extension_penalty
        self.auto_adjust_extension_penalty = auto_adjust_extension_penalty
        self.no_realignment = no_realignment
        self.is_complex_input = is_complex_input
        self.second_target = second_target
        self.unspliced_local_reference = UnsplicedLocalReference(
                                            self.__target.chrom,
                                            self.__target.pos,
                                            self.__target.reference.get_reference_length(self.__target.chrom),
                                            self.window,
                                            self.__target.reference
                                         )
        self.__pileup, self.contig = self.__parse_pileup()

    cdef __parse_pileup(self, Contig contig=None, bint retargeted=False, bint skip_read_end_check=False):
        """Dictize reads for target indel and make target indel template by consensus

        bam (pysam.AlignmentFile)
        target_indel (Variant): indel of interest
        smith_waterman (bool): True to perform SW realn
        template (bool): None if no indel template supplied
        """

        cdef list pileup

        read_end_evidence_only = False

        # equivalence search
        if retargeted:
            pileup = self.__pileup
            sample_factor = self.__sample_factor
        else:
            pileup, self.__sample_factor = make_pileup(
                self.__target,
                self.bam,
                self.unspliced_local_reference,
                exclude_duplicates=self.exclude_duplicates,
                window=self.window,
                downsamplethresh=self.downsamplethresh,
                basequalthresh=self.basequalthresh,
            )

            (
            self.__target,
            pileup,
            exptension_penalty_used,
            self._observed_pos,
            read_end_evidence_only
            ) = find_by_normalization(
                self.__target,
                pileup,
                self.window,
                self.match_score,
                self.mismatch_penalty,
                self.gap_open_penalty,
                self.gap_extension_penalty,
                self.basequalthresh,
            )

            if skip_read_end_check:
                read_end_evidence_only = False

            if self.target != self.__target:
                    self.__target, pileup = update_pileup(
                        pileup,
                        self.__target,
                        self.window,
                        self.match_score,
                        self.mismatch_penalty,
                        self.gap_open_penalty,
                        self.gap_extension_penalty,
                        self.basequalthresh,
                        bypass_search=True
                    )

            contig = Contig(
                self.__target,
                preprocess_for_contig_construction(
                    self.__target,
                    self.target,
                    pileup,
                    self.unspliced_local_reference,
                    self.window,
                    self.match_score,
                    self.mismatch_penalty,
                    self.gap_open_penalty,
                    exptension_penalty_used,
                ),
                self.unspliced_local_reference,
                self.basequalthresh,
                self.mapqthresh,
            )

            self.is_spurious_overhang = False
            if contig.failed and not self.no_realignment:
                within = self.retarget_window

                grid = generate_grid(self.auto_adjust_extension_penalty,
                                     self.gap_open_penalty,
                                     self.gap_extension_penalty,
                                     self.__target,
                       )

                ans = check_overhangs(pileup)

                if ans:
                    intron, overhangs = ans[0], ans[1]
                    non_spurious_overhangs = filter_spurious_overhangs(
                        self.__target,
                        intron,
                        overhangs,
                        self.match_score,
                        self.mismatch_penalty,
                        self.gap_open_penalty,
                        self.gap_extension_penalty,
                    )
                    if not non_spurious_overhangs:
                        contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
                        self.is_spurious_overhang = True
                        return pileup, contig
                    else:
                        res = grid_search(
                            self.__target,
                            non_spurious_overhangs,
                            self.window,
                            self.mapqthresh,
                            within,
                            self.retarget_cutoff,
                            self.match_score,
                            self.mismatch_penalty,
                            grid,
                            self.unspliced_local_reference,
                            self.exact_match_for_shiftable,
                        )

                        if res:
                            self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
                        else:
                            contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
                            self.is_spurious_overhang = True
                            return pileup, contig
                else:
                    res = grid_search(
                        self.__target,
                        pileup,
                        self.window,
                        self.mapqthresh,
                        within,
                        self.retarget_cutoff,
                        self.match_score,
                        self.mismatch_penalty,
                        grid,
                        self.unspliced_local_reference,
                        self.exact_match_for_shiftable,
                    )

                    if res:
                        self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]

                # if retargeted successfully -> make contig based on the retarget
                if res:
                    self.__target, retarget_reads = res[0], res[1]

                    self.__target, self.__pileup = update_pileup(
                        pileup,
                        self.__target,
                        self.window,
                        self.match_score,
                        self.mismatch_penalty,
                        self.gap_open_penalty,
                        self.gap_extension_penalty,
                        self.basequalthresh,
                        bypass_search=True,
                    )

                    contig = Contig(
                        self.__target,
                        preprocess_for_contig_construction(
                            self.__target,
                            self.__target,
                            self.__pileup,
                            self.unspliced_local_reference,
                            self.window,
                            self.match_score,
                            self.mismatch_penalty,
                            self.gap_open_penalty,
                            self.gap_extension_penalty,
                        ),
                        self.unspliced_local_reference,
                        self.basequalthresh,
                        self.mapqthresh
                    )

                    # 2nd-pass using the retarget
                    return self.__parse_pileup(contig=contig, retargeted=True)

                # no target in this pileup
                else:
                    if self.is_complex_input:
                        try:
                            self.__target = self.second_target
                            self.target = self.second_target
                            self.is_complex_input = False
                            return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
                        except:
                             pileup, contig
                    else:
                        return pileup, contig

        #_target = [read for read in pileup if read["is_target"]]
        #_nontarget = [read for read in pileup if not read["is_target"]]
        #_target = sorted(_target, key=partial(centrality, target_pos=self.__target.pos))
        #_nontarget = [read for read in pileup if not read["is_target"]]
        #_nontarget = sorted(_nontarget, key=partial(centrality, target_pos=self.__target.pos))

        #print(_target[0])
        #print(_nontarget[0])

        # soft-clip realn & SW realn
        if contig.qc_passed and not self.no_realignment:

            orig_contig = contig

            # realign reads that are not through retarget path
            if not retargeted:
               cutoff = 1.0
               within = 30

               target = [read for read in pileup if read["is_target"]]
               nontarget = [read for read in pileup if not read["is_target"]]

               grid = generate_grid(self.auto_adjust_extension_penalty,
                                    self.gap_open_penalty,
                                    self.gap_extension_penalty,
                                    self.__target,
                      )

               res = grid_search(
                       self.__target,
                       nontarget,
                       self.window,
                       self.mapqthresh,
                       within,
                       cutoff,
                       self.match_score,
                       self.mismatch_penalty,
                       grid,
                       self.unspliced_local_reference,
                       self.exact_match_for_shiftable,
                   )

               if res:
                   nontarget = [read for read in nontarget if read not in res[1]]

                   pileup = target + res[1] + nontarget
                   self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]

                   self.__target, pileup = update_pileup(
                       pileup,
                       self.__target,
                       self.window,
                       self.match_score,
                       self.mismatch_penalty,
                       self.gap_open_penalty,
                       self.gap_extension_penalty,
                       self.basequalthresh,
                       bypass_search=True,
                   )

                   #equivalent but different position
                   if self.__target == res[0]:
                       self.__target = res[0]

               else:
                  pileup = target + nontarget

            if self.__target.count_repeats() == 0:
                pileup = find_by_softclip_split(self.__target, contig, pileup)


            if read_end_evidence_only:
                target_pileup = [read for read in pileup if read["is_target"]]

            pileup = find_by_smith_waterman_realn(
                self.__target,
                contig,
                pileup,
                self.match_score,
                self.mismatch_penalty,
                self.gap_open_penalty,
                self.gap_extension_penalty,
                self.basequalthresh
            )

            if read_end_evidence_only:
                newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]
                if newly_identified:
                    expected_ptrn = most_common_gap_ptrn(newly_identified)

                    indels = []
                    contig_seq = contig.get_contig_seq()
                    alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
                    for new_one in newly_identified:
                        if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):
                            indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]

                    if indels:
                        try:
                            self.__target = most_common(indels)
                        except:
                            target_pos = self.__target.pos
                            indels.sort(key=lambda x : abs(x.pos - target_pos))

                        return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)

            contig = Contig(
                self.__target,
                preprocess_for_contig_construction(
                    self.__target,
                    self.target,
                    pileup,
                    self.unspliced_local_reference,
                    self.window,
                    self.match_score,
                    self.mismatch_penalty,
                    self.gap_open_penalty,
                    self.gap_extension_penalty,
                ),
                self.unspliced_local_reference,
                self.basequalthresh,
                self.mapqthresh,
            )

            contig = compare_contigs(orig_contig, contig, self.__target.pos)

        return pileup, contig


    def __eq__(self, other):
        # alignment equivalence
        my_contig, other_contig = self.contig, other.contig

        if my_contig.failed or other_contig.failed:
            return False

        # check eq in phased form
        phasing_mode = "local"
        my_phased, other_phased = self.phase(how=phasing_mode), other.phase(how=phasing_mode)

        return (my_phased == other_phased)


    def __hash__(self):
        hashable = self.phase(how="local")
        return hash(hashable)


    def get_contig(self):
        """returns :class:`~indelpost.Contig` object built from reads supporting the target indel.
        :class:`~indelpost.FailedContig` is returned if contig assembly is not successful.
        """
        contig = self.contig
        if contig and not contig.failed:
            return contig
        else:
            failed = FailedContig()
            alt_cnt = self.count_alleles()[1]
            if alt_cnt:

                dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]
                if sum(dirty_target_pileup) == len(dirty_target_pileup):
                    failed.is_low_quality = True
                else:
                    failed.failed_anyway = True
            else:
                failed.target_not_found = True

            return failed


    def get_target_indel(self):
        """returns :class:`~indelpost.Variant` object representing the actual target indel analyzed.
        The target indel may be different from the input indel due to the internal realignment.
        :class:`~indelpost.NullVariant` is returned when no target is found.
        """
        alt_cnt = self.count_alleles()[1]
        if alt_cnt:
            return self.__target
        else:
            return NullVariant(self.__target.chrom, self.__target.pos, self.__target.reference)


    def fetch_reads(self, how='target'):
        """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__
        of `pysam.AlignedSegment <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment>`__ objects,
        depending on the strategy specified.

        Parameters
        ----------
            how : string
                specifies the read fetch strategy. "target" (default) to retrieve reads supporting the target.
                "non_target" for reads w/o target. "covering" for reads covering the locus w/ and w/o the target.
        """

        if how == "target":
            return [read["read"] for read in self.__pileup if read["is_target"]]
        elif how == "non_target":
            pos, indel_len = self._observed_pos, len(self.target.indel_seq)
            r_pos = max(v.pos for v in self.target.generate_equivalents())
            margin = r_pos - pos
            del_len = indel_len if self.target.is_del else 0
            targets = [read["read_name"] for read in self.__pileup if read["is_target"]]
            return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]
        elif how == "covering":
            return [read["read"] for read in self.__pileup if read["is_covering"]]
        else:
            raise Exception("fetch stragety must be either of target, non_target, covering")


    def count_alleles(
        self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None
    ):
        """returns a `tuple <https://docs.python.org/3/library/stdtypes.html#tuple>`__ of
        read counts:  (#non_target reads, #target reads).

        Parameters
        ----------
            fwrv : bool
                breaks down to the forward and reverse read counts.
                ( (non_target-fw, non_target-rv), (target-fw, target-rv) ).
            by_fragment : bool
                counts by fragment. Overlapping fw and rv reads are counted as one.
            three_class : bool
                breaks down to (reference, non_reference_non_target, target) counts.
            estimated_count : bool
                True to return estimated count when the coverage is higher than :attr:`~indelpost.VariantAlignment.downsample_threshold`.
            quality_window : integer
                specifies the range of base call quality filter. indel pos +/- quality_window.
            quality_threshold : integer
                filters reads with the median base call quality in indel pos +/- quality_window < quality_threshold.
        """

        cdef dict read

        pos, indel_len = self._observed_pos, len(self.target.indel_seq)

        r_pos = max(v.pos for v in self.__target.generate_equivalents())

        margin = r_pos - pos

        del_len = indel_len if self.target.is_del else 0

        reads = self.__pileup
        if quality_window and quality_threshold:
            reads = [
                read
                for read in reads
                if is_quality_read(read, pos, quality_window, quality_threshold)
            ]

        if three_class:
            for read in reads:
                read["is_locally_ref"] = is_locally_ref(read, pos)

        fw_target = [
            read["read_name"]
            for read in reads
            if read["is_target"] and not read["is_reverse"]
        ]
        rv_target = [
            read["read_name"]
            for read in reads
            if read["is_target"] and read["is_reverse"]
        ]

        fw_non_target = [
            read["read_name"]
            for read in reads
            if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]
        ]
        rv_non_target = [
            read["read_name"]
            for read in reads
            if count_as_non_target(read, pos, del_len, margin) and read["is_reverse"]
        ]

        fw_target = set(fw_target)
        rw_target = set(rv_target)
        fwrv_target = fw_target | rw_target

        fw_non_target = set(fw_non_target) - fwrv_target
        rv_non_target = set(rv_non_target) - fwrv_target

        est = self.__sample_factor if estimated_count else 1

        if three_class:
            fw_ref = [
                read["read_name"]
                for read in reads if read["is_locally_ref"] and not read["is_reverse"] and read["read_name"] in fw_non_target
            ]

            fw_ref = set(fw_ref)
            fw_non_ref_non_target = fw_non_target - fw_ref

            rv_ref = [
                read["read_name"]
                for read in reads if read["is_locally_ref"] and read["is_reverse"] and read["read_name"] in rv_non_target
            ]
            rv_ref = set(rv_ref)
            rv_non_ref_non_target = rv_non_target - rv_ref

            if fwrv:
                return (
                    (
                        int(len(fw_ref) * est),
                        int(len(rv_ref) * est),
                    ),
                    (
                        int(len(fw_non_ref_non_target) * est),
                        int(len(rv_non_ref_non_target) * est),
                    ),
                    (
                        int(len(fw_target) * est),
                        int(len(rv_target) * est),
                    ),
                )

            else:
                if by_fragment:
                    fwrv_ref = len(fw_ref | rv_ref)
                    fwrv_non_ref_non_target = len(fw_non_ref_non_target | rv_non_ref_non_target)
                    fwrv_target = len(fw_target | rw_target)
                else:
                    fwrv_ref = len(fw_ref) + len(rv_ref)
                    fwrv_non_ref_non_target =  len(fw_non_ref_non_target) + len(rv_non_ref_non_target)
                    fwrv_target = len(fw_target) + len(rw_target)

                return (int(fwrv_ref), int(fwrv_non_ref_non_target), int(fwrv_target))

        if fwrv:
            return (
                (
                    int(len(fw_non_target) * est),
                    int(len(rv_non_target) * est),
                ),
                (
                    int(len(fw_target) * est),
                    int(len(rv_target) * est),
                ),
            )
        else:
            if by_fragment:
                fwrv_non_target = len(fw_non_target | rv_non_target)
                fwrv_target = len(fw_target | rw_target)
            else:
                fwrv_non_target = len(fw_non_target) + len(rv_non_target)
                fwrv_target = len(fw_target) + len(rw_target)

            return (
                int(fwrv_non_target * est),
                int(fwrv_target * est),
            )


    def phase(
        self,
        how="local",
        local_threshold=20,
        longest_common_substring_threshold=15,
        indel_repeat_threshold=None,
        mutation_density_threshold=0.05,
    ):
        """returns a :class:`~indelpost.Variant` object represeting a phased target indel.
        :class:`~indelpost.NullVariant` is returned if no target is found.
        Phasing is limited to the exon where the target indel is located for RNA-Seq.
        Refer to :meth:`~indelpost.Contig.get_phasables()` to retrieve phasable variants across exons.

        Parameters
        ----------
        how : string
            - "local" (default) phasing by recursive elimination of longest common substrings (LCS) and locality  scoring (see local_threshold).
            - "greedy" phase all phasable events.
            - "complex" phasing by the "local" option + exclusivity check. The exclusivity check removes phasable events that are also observed in non-target reads such as nearby homozygous germline events.
        local_threshold : integer
            local (and complex) phasing method checks if phasable SNVs are locally clustered around the indel.
            For i-th base apart from indel, the score gains one if mismatch and, if match, loses
            1*min(j/local_thresh, 1) where j = i for short indels (< 10-nt) and 0.6*i for longer indels.
        longest_common_substring_threshold : integer
            removes common substrings between the reference and contig assembled from target reads that are longer than longest_common_substring_threshold (default 15).
        indel_repeat_threshold : integer
            do not phase indels that repeat more than indel_repeat_threshold (default None)
        mutation_density_threshold : float
            do not phase if the pileup contains too many mutations (possibly error-prone dirty region).
            In non-target reads, #non-ref bases/#all bases > mutation_density_threshold (default 0.05)
        """
        if how == "complex":
            hard, to_complex = False, True
        elif how == "greedy":
            hard, to_complex = True, False
        elif how == "local":
            hard, to_complex = False, False
        else:
            raise Exception("phasing stragety must be either of local, greedy, complex")

        if indel_repeat_threshold is None:
            indel_repeat_threshold = np.inf

        return phase_nearby_variants(
            self.__target,
            self.contig,
            self.__pileup,
            self.basequalthresh,
            local_threshold,
            longest_common_substring_threshold,
            indel_repeat_threshold,
            mutation_density_threshold,
            hard,
            to_complex
        )


def is_quality_read(read, pos, qualitywindow, qualitythresh):

    try:
        lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]
    except:
        lt_qual, rt_qual = split(
            read["read_qual"],
            read["cigar_string"],
            pos,
            read["read_start"],
            is_for_ref=False,
            reverse=False,
        )

    if lt_qual and rt_qual:
        lt_median = np.median(lt_qual[-min(len(lt_qual), qualitywindow) :])
        rt_median = np.median(rt_qual[: min(len(rt_qual), qualitywindow)])

        return lt_median > qualitythresh and rt_median > qualitythresh


def is_locally_ref(read, pos):
    if read["is_reference_seq"]:
        return True


    try:
        lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
    except:
        lt_seq, rt_seq = split(read["read_seq"],
            read["cigar_string"],
            pos,
            read["read_start"],
            is_for_ref=False,
            reverse=False,
        )

    try:
        lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
    except:
        lt_ref, rt_ref = split(read["ref_seq"],
            read["cigar_string"],
            pos,
            read["aln_start"],
            is_for_ref=True,
            reverse=False,
        )



    lt_seq_len = len(lt_seq)
    lt_ref_len = len(lt_ref)
    #clipped
    if not lt_ref_len:
        return False

    lt_len = min(5, lt_seq_len, lt_ref_len)

    rt_seq_len = len(rt_seq)
    rt_ref_len = len(rt_ref)
    if not rt_ref_len:
        return False

    rt_len = min(5, rt_seq_len, rt_ref_len)

    if lt_seq[-lt_len:] == lt_ref[-lt_len:] and rt_seq[:rt_len] == rt_ref[:rt_len]:
        return True
    else:
        return False


cdef bint count_as_non_target(dict read, int pos, int del_len, int margin):
    if read["is_target"]:
        return False

    cdef int aln_start = read["aln_start"]
    cdef int aln_end = read["aln_end"]

    # undetermined reads
    if read.get("undetermined", False):
        return False

    if read["is_covering"]:
        covering_subread = read["covering_subread"]
        if covering_subread[1] <= pos + margin:
            return False

        if pos < aln_start or aln_end < pos:
            return False
    else:
        if aln_end < pos:
            return False

        if del_len:
            if pos + del_len < aln_start:
                return False
        else:
            return False

    return True

def centrality(read, target_pos):
    relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)
    return abs(0.5 - relative_pos)

cdef list preprocess_for_contig_construction(
    Variant target,
    Variant orig_target,
    list pileup,
    UnsplicedLocalReference unspl_loc_ref,
    int window,
    int match_score,
    int mismatch_penalty,
    int gap_open_penalty,
    int gap_extension_penalty,
):

    cdef dict read
    cdef int clips, nonclips

    if not pileup:
        return pileup

    targetpileup = [read for read in pileup if read["is_target"] and not read["is_dirty"]]

    if not targetpileup:
        return targetpileup

    clipped_targetpileup = [
        read for read in targetpileup if "S" in read["cigar_string"]
    ]
    nonclipped_targetpileup = [
        read for read in targetpileup if not "S" in read["cigar_string"]
        and (read.get("lt_cigar", None) and read.get("rt_cigar", None))
    ]

    clips = len(clipped_targetpileup)
    nonclips = len(nonclipped_targetpileup)

    if target == orig_target and nonclips > 9:
        random.seed(123)
        targetpileup = random.sample(nonclipped_targetpileup, 10)
        targetpileup = [right_aligner(read, target) for read in targetpileup]
    else:
        targetpileup = sorted(targetpileup, key=partial(centrality, target_pos=target.pos))

        unspl_ref_seq, unspl_lt_len = get_local_reference(orig_target, pileup, window, unspl_loc_ref, unspliced=True)
        unspl_aligner = make_aligner(unspl_ref_seq, match_score, mismatch_penalty)
        unspl_start = orig_target.pos + 1 - unspl_lt_len

        is_gapped_aln = False
        targetpileup = [
            update_spliced_read_info(
                read,
                target,
                orig_target,
                is_gapped_aln,
                window,
                match_score,
                mismatch_penalty,
                gap_open_penalty,
                gap_extension_penalty,
                unspl_loc_ref,
            )
            if "N" in read["cigar_string"] else
            update_read_info(
                read,
                target,
                is_gapped_aln,
                gap_open_penalty,
                gap_extension_penalty,
                unspl_aligner,
                unspl_ref_seq,
                unspl_start,
            )
            for read in targetpileup
        ]

        targetpileup = [read for read in targetpileup if read is not None and (read.get("lt_cigar", None) and read.get("rt_cigar", None))]

        _targetpileup = [
            read for read in targetpileup if read.get("cigar_updated", False)
        ]

        if _targetpileup:
            targetpileup = _targetpileup
        else:
            return targetpileup

    return targetpileup


def update_spliced_read_info(
    read,
    target,
    orig_target,
    is_gapped_aln,
    window,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    unspl_loc_ref,
):
    ref_seq, lt_len = get_local_reference(orig_target, [read], window, unspl_loc_ref)
    aligner = make_aligner(ref_seq, match_score, mismatch_penalty)
    ref_start = orig_target.pos + 1 - lt_len

    read = update_read_info(
        read,
        target,
        is_gapped_aln,
        gap_open_penalty,
        gap_extension_penalty,
        aligner,
        ref_seq,
        ref_start
    )

    return right_aligner(read, target)

def right_aligner(read, target):
    """Right align indels around splice site"""

    if (
        "N" not in read["cigar_string"]
        or (
            "I" not in read["cigar_string"]
            and "D" not in read["cigar_string"]
        )
    ):
        return read

    cigar_lst = read["cigar_list"]

    query_pos = 0
    ref_pos = read["aln_start"]
    new_cigar = []
    prev_event = "A"
    skip_next = False
    for i, c in enumerate(cigar_lst):
        event, event_len = c[-1], int(c[:-1])

        if event_len < 0:
            return None

        query_move = 0 if event in ("D", "N", "H", "P") else event_len
        ref_move = 0 if event in ("I", "H", "P") else event_len

        if event in ("I", "D") and prev_event == "N":
            try:
                nxt_c = cigar_lst[i + 1]
                nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
                if nxt_event != "M":
                    raise Exception
            except:
                return None

            chrom, reference = target.chrom, target.reference
            padding_base = reference.fetch(chrom, ref_pos - 2, ref_pos - 1)
            if event == "I":
                ins_seq = read["read_seq"][query_pos: query_pos + event_len]
                ref = padding_base
                alt = padding_base + ins_seq
            else:
                del_seq = reference.fetch(chrom, ref_pos - 1, ref_pos - 1 + event_len)
                ref = padding_base + del_seq
                alt = padding_base

            right_aligned_vars = Variant(
                                    chrom,
                                    ref_pos - 1,
                                    ref,
                                    alt,
                                    reference,
                                    skip_validation=True
                                ).generate_equivalents()

            diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1
            if diff > 0:
                new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
            else:
                return None

            ref_pos += query_move + nxt_event_len
            query_pos += ref_move + nxt_event_len
            skip_next = True

        else:
            if skip_next:
                skip_next = False
            else:
                query_pos += query_move
                ref_pos += ref_move
                new_cigar.append(c)

        prev_event = event

    read["cigar_list"] = new_cigar
    read["cigar_string"] = "".join(new_cigar)

    try:
        if target in right_aligned_vars:
            rt_aln_pos = target.pos + diff
            read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])
            read["lt_flank"], read["rt_flank"] = split(
                                                      read["read_seq"],
                                                      read["cigar_string"],
                                                      rt_aln_pos,
                                                      read["read_start"],
                                                      is_for_ref=False,
                                                      reverse=False
                                                 )
            read["lt_qual"], read["rt_qual"] = split(
                                                    read["read_qual"],
                                                    read["cigar_string"],
                                                    rt_aln_pos,
                                                    read["read_start"],
                                                    is_for_ref=False,
                                                    reverse=False
                                               )
            read["lt_ref"], read["rt_ref"] = split(
                                                    read["ref_seq"],
                                                    read["cigar_string"],
                                                    rt_aln_pos, read["aln_start"],
                                                    is_for_ref=True,
                                                    reverse=False
                                             )
            read["target_right_shifted"] = rt_aln_pos

            indel_len = len(target.indel_seq)
            if target.is_ins:
                read["rt_flank"] = read["rt_flank"][indel_len :]
                read["rt_qual"] = read["rt_qual"][indel_len :]
            else:
                read["rt_ref"] = read["rt_ref"][indel_len :]
        else:
            read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], target.pos, read["read_start"])
    except:
        pass

    return read


def generate_grid (auto_adjust_extension_penalty,
                   gap_open_penalty,
                   gap_extension_penalty,
                   target,
):
    if auto_adjust_extension_penalty:
        if (gap_open_penalty, gap_extension_penalty) != (3, 1):
            if len(target.indel_seq) < 20:
                grid = [(gap_open_penalty, gap_extension_penalty),
                            (3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)
                       ]
            else:
                grid = [(gap_open_penalty, gap_extension_penalty),
                            (3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)
                       ]
        else:
            if len(target.indel_seq) < 20:
                grid = [(3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)]
            else:
                grid = [(3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)]
    else:
        grid = [(gap_open_penalty, gap_extension_penalty)]

    return grid


def grid_search(
    target,
    pileup,
    window,
    mapq_thresh,
    within,
    retarget_cutoff,
    match_score,
    mismatch_penalty,
    grid,
    unspl_loc_ref,
    exact_match_for_shiftable,
):
    # grid = [(gap.open, gap.ext)]
    h = 0
    responses, scores, hs = [], [], []
    while h < len(grid):
        res = retarget(
            target,
            pileup,
            window,
            mapq_thresh,
            within,
            retarget_cutoff,
            match_score,
            mismatch_penalty,
            grid[h][0],
            grid[h][1],
            unspl_loc_ref,
            exact_match_for_shiftable,
        )

        if res:
            score = res[2]
            responses.append(res)
            #scores.append(score)
            hs.append(h)

            # exact match
            if score == 1.0:
                scores.append(score * len(res[1]))
                # cnt exact hit?
                #break
            else:
                scores.append(score)
        h += 1

    if responses:
        idx = scores.index(max(scores))
        best_res = responses[idx]
        best_params = grid[hs[idx]]

        is_gapped_aln=False # to be removed

        candidate = best_res[0]

        gap_open_penalty, gap_extension_penalty = best_params[0], best_params[1]

        updated_reads = []
        for read, aligner, ref_seq, ref_start in zip(best_res[1], best_res[5], best_res[3], best_res[4]):

            updated_read = update_read_info(
                                read,
                                candidate,
                                is_gapped_aln,
                                gap_open_penalty,
                                gap_extension_penalty,
                                aligner,
                                ref_seq,
                                ref_start,
                            )

            updated_reads.append(updated_read)


        return candidate, updated_reads, gap_open_penalty, gap_extension_penalty
    else:
        return None


def is_perfect_match(aligner, contig_seq, read_seq):
    aligner.setRead(read_seq)
    _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
    _contig = contig_seq[_aln.reference_start : _aln.reference_end]
    _read = read_seq[_aln.read_start : _aln.read_end]

    return _contig == _read





# 786
# Aldy source: conftest.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest

from aldy.gene import Gene
from aldy.common import script_path


@pytest.fixture
def toy_gene():
    return Gene(script_path("aldy.tests.resources/toy.yml"))


@pytest.fixture
def real_gene():
    return Gene(script_path("aldy.resources.genes/cyp2d6.yml"))


def pytest_addoption(parser):
    parser.addoption("--solvers", action="store", default=None)


def pytest_generate_tests(metafunc):
    solvers = metafunc.config.getoption("solvers")
    if solvers is None:
        solvers = "any"
    solvers = solvers.split(",")

    if "solver" in metafunc.fixturenames:
        metafunc.parametrize("solver", solvers)



# cython: embedsignature=True

from pysam.libcfaidx cimport FastaFile
from pysam.libcbcf cimport VariantRecord, VariantFile

cdef class NullVariant:
    cdef readonly str chrom
    cdef readonly str ref, alt
    cdef readonly int pos
    cdef readonly FastaFile reference


cdef class Variant:
    cdef public str chrom, ref, alt,
    cdef public int pos
    cdef public FastaFile reference
    cdef str _chrom
    cdef VariantFile vcf
    cdef VariantRecord rec, hit

    cpdef __validate(self)




#!/usr/bin/env python
# 786
# Aldy source: __main__.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Optional, Any
import logbook
import logbook.more
import logbook.base
import argparse
import os
import sys
import platform
import yaml
import datetime
import tempfile
import traceback
import pytest

from . import common
from .common import log, script_path, AldyException, td, parse_cn_region
from .gene import Gene
from .profile import Profile
from .genotype import genotype
from .query import query
from .version import __version__


def get_version():
    return "{} {}".format(
        platform.system() if platform.system() != "Darwin" else "macOS",
        (
            platform.platform()[6:]
            if platform.system() == "Linux"
            else (
                platform.mac_ver()[0]
                if platform.system() == "Darwin"
                else platform.platform()
            )
        ),
    )


def main(argv):
    """
    The main entry point.
    """

    parser, args = _get_args(argv)

    # Set the logging verbosity
    level = (args.verbosity if "verbosity" in args else "I").lower()
    level = next(
        v
        for k, v in logbook.base._reverse_level_names.items()
        if k.lower().startswith(level)
    )

    # Set the command-line logging
    sh = logbook.more.ColorizedStderrHandler(
        format_string="{record.message}", level=level
    )
    sh.push_application()

    log.info(
        "🐿  Aldy v{} (Python {} on {})",
        __version__,
        platform.python_version(),
        get_version(),
    )
    log.info(
        "   (c) 2016-{} Aldy Authors. All rights reserved.\n"
        + "   Free for non-commercial/academic use only.",
        datetime.datetime.now().year,
    )

    try:
        if not args.subparser or args.subparser == "help":
            parser.print_help()
        elif args.subparser == "license":
            _print_licence()
        elif args.subparser == "test":
            _run_test()
        elif args.subparser in ["query", "q"]:
            q = args.query
            if "*" in q:
                gene, q = q.split("*", maxsplit=1)
            else:
                gene, q = q, ""
            if args.gene:
                gene = args.gene
            db_file = script_path("aldy.resources.genes/{}.yml".format(gene.lower()))
            if not os.path.exists(db_file):
                db_file = gene
            with open(db_file):  # Check if file exists
                pass
            query(Gene(db_file, genome=args.genome), q)
        elif args.subparser == "profile":
            params = {}
            if args.param:
                for pl in args.param:
                    for p in pl:
                        if "=" not in p:
                            raise AldyException(f"Invalid parameter {p}")
                        k, v = p.split("=", 1)
                        params[k.replace("-", "_")] = v
            try:
                p = Profile.get_sam_profile_data(
                    args.file,
                    cn_region=parse_cn_region(args.cn_neutral_region),
                    genome=args.genome,
                    params=params,
                )
                print(yaml.dump(p, default_flow_style=None))
            except AldyException as ex:
                log.error("ERROR: {}", str(ex))
        elif args.subparser in ["genotype", "g"]:
            # Prepare the output file
            output = args.output
            if output == "-":
                output = sys.stdout
            elif output:
                output = open(output, "w")
            _genotype(args.gene, output, args)
            if output and output != sys.stdout:
                output.close()
        else:
            raise AldyException("Invalid sub-command " + args.subparser)
    except IOError as ex:
        if ex.filename is not None:
            log.critical("ERROR: {} cannot be accessed", ex.filename)
        else:
            log.critical("ERROR: {} cannot be accessed", str(ex))
        log.debug(ex)
        log.debug(traceback.format_exc())
        exit(1)
    except SystemExit as ex:
        log.debug(repr(ex))
        log.debug(traceback.format_exc())
        exit(ex.code)
    except Exception as ex:
        gene = "" if "gene" not in args else f"gene= {args.gene}, "
        log.critical(f"ERROR: {gene}file= {args.file if 'file' in args else '-'}")
        log.critical(repr(ex))
        log.warn(traceback.format_exc())
        exit(1)
    except:  # noqa
        gene = "" if "gene" not in args else f"gene= {args.gene}, "
        exc = sys.exc_info()[0]
        log.critical(f"ERROR: {gene}file= {args.file if 'file' in args else '-'}")
        log.critical("Unrecoverable error: {}", repr(exc))
        log.warn(traceback.format_exc())
        exit(1)


def _get_args(argv):
    """
    Parse command-line arguments.
    """

    parser = argparse.ArgumentParser(
        prog="aldy",
        description="Allelic decomposition and exact genotyping of highly polymorphic "
        + "and structurally variant genes",
    )

    base = argparse.ArgumentParser(add_help=False)
    base.add_argument(
        "--verbosity",
        "-v",
        default="INFO",
        help=td(
            """Logging verbosity:
               - T (trace)
               - D (debug)
               - I (info) and
               - W (warn).
               Default is "I" (info)."""
        ),
    )

    subparsers = parser.add_subparsers(dest="subparser")

    genotype_parser = subparsers.add_parser(
        "genotype",
        help="Call the most likely genotype and diplotype within a sample.",
        parents=[base],
    )
    genotype_parser.add_argument(
        "file", nargs="?", help="Input file in SAM, BAM or CRAM format."
    )
    genotype_parser.add_argument(
        "--gene",
        "-g",
        default="all",
        help='Gene whose genotype is to be called. Default is "all" which calls '
        + "genotypes for all supported genes.",
    )
    genotype_parser.add_argument(
        "--profile",
        "-p",
        help=td(
            """Sequencing profile. The following profiles are supported:
               - illumina (also: wgs),
               - pgrnseq-v1 (also: pgx1),
               - pgrnseq-v2 (also: pgx2),
               - pgrnseq-v3 (also: pgx3),
               - 10x,
               - pacbio-hifi-targeted,
               - pacbio-hifi-targeted-twist,
               - exome (also: wxs, wes).
               You can also provide a custom SAM/BAM file as a profile.
               Please check documentation for more details."""
        ),
    )
    genotype_parser.add_argument(
        "--reference",
        "-r",
        default=None,
        help="Genome reference used for reading CRAM files",
    )
    genotype_parser.add_argument(
        "--genome",
        default=None,
        help="SAM/BAM reference genome (hg19 or hg38; none for auto-detection)",
    )
    genotype_parser.add_argument(
        "--cn-neutral-region",
        "-n",
        help=td(
            """Copy-number neutral region in the format chromosome:start-end
               (e.g. chr1:10000-20000).
               Default is CYP2D8 region within hg19 (22:42547463-42548249)."""
        ),
    )
    genotype_parser.add_argument(
        "--output",
        "-o",
        default=None,
        help="Output file location. Default is [input].[gene].aldy.",
    )
    genotype_parser.add_argument(
        "--solver",
        "-s",
        default="any",
        help=td(
            """ILP solver:
               - gurobi (Gurobi)
               - cbc (Google OR-Tools/CBC)
               - any (attempts to use Gurobi and then CBC).
               Default is "any"."""
        ),
    )
    genotype_parser.add_argument(
        "--debug",
        default=None,
        help="Create a directory that will contain the debug information "
        + "and core dumps.",
    )
    genotype_parser.add_argument(
        "--cn",
        "-c",
        default=None,
        help=td(
            """Manually set the copy number configuration as a list of comma-separated
               configuration IDs (e.g. "CN1,CN2").
               For a list of supported configuration IDs, please run `aldy show`.
               For the most common diploid case that contains no structural variations
               (e.g. two copies of the main gene), use 1,1."""
        ),
    )
    genotype_parser.add_argument("--log", "-l", default=None, help="Log file location")
    genotype_parser.add_argument(
        "--multiple-warn-level",
        "-W",
        default=1,
        type=int,
        help="Show warning if multiple solutions are found. "
        + "Can be 1 (warn after genotyping) or 2 "
        + "(also warn if there are multiple major solutions)."
        + "Default is 1 (warn after the genotyping).",
    )
    genotype_parser.add_argument(
        "--simple",
        action="store_true",
        default=False,
        help=td("""Print one-line result per gene (debug). Default is off."""),
    )
    genotype_parser.add_argument("--param", action="append", nargs="+")

    _ = subparsers.add_parser(
        "test",
        parents=[base],
        help="Run Aldy test suite. Recommended prior to the first use",
    )

    _ = subparsers.add_parser("license", parents=[base], help="Show Aldy license")

    show_parser = subparsers.add_parser(
        "query",
        aliases=["q"],
        parents=[base],
        help="Query database definitions for a given gene.",
    )
    show_parser.add_argument("--gene", "-g", default=None, help="Gene file.")
    show_parser.add_argument("--genome", default="hg19", help="Reference genome.")
    show_parser.add_argument("query", help="Gene or allele to show.")

    profile_parser = subparsers.add_parser(
        "profile",
        parents=[base],
        help=td(
            """Generate a sequencing profile for a SAM/BAM/CRAM file.
               Please check the documentation for more details."""
        ),
    )
    profile_parser.add_argument("file", nargs="?", help="SAM/BAM/CRAM file.")
    profile_parser.add_argument(
        "--cn-neutral-region",
        "-n",
        default=None,
        help=td(
            """Copy-number neutral region in the format chromosome:start-end
               (e.g. chr1:10000-20000).
               Default is CYP2D8 region within hg19 (22:42547463-42548249)."""
        ),
    )
    profile_parser.add_argument(
        "--genome",
        default=None,
        help="SAM/BAM reference genome (hg19 or hg38; hg19 by default)",
    )
    profile_parser.add_argument("--param", action="append", nargs="+")

    _ = subparsers.add_parser(
        "help", parents=[base], help="Show program usage and exit."
    )

    return parser, parser.parse_args(argv)


def _print_licence():
    """
    Print Aldy license.
    """
    with open(script_path("aldy.resources/LICENSE.rst")) as f:
        for ll in f:
            print(ll.strip())


def _genotype(gene: str, output: Optional[Any], args) -> None:
    """
    Genotype a file.

    :raise: :py:class:`aldy.common.AldyException` if `cn_region` is invalid.
    """

    cn_region = parse_cn_region(args.cn_neutral_region)
    cn_solution = args.cn
    if cn_solution:
        cn_solution = cn_solution.split(",")

    def run(debug):
        log.trace(
            "\n[main] arguments= {}",
            " ".join(k + "=" + str(v) for k, v in vars(args).items() if k is not None),
        )
        log.info("Genotyping sample {}...", os.path.basename(args.file))

        if args.profile in ["exome", "wxs"]:
            log.warn("WARNING: Copy-number calling is not available for exome data.")
            log.warn(
                "WARNING: Aldy will NOT be able to detect gene duplications, "
                + "deletions and fusions."
            )
            log.warn(
                "WARNING: Calling of alleles that are defined by non-exonic mutations "
                + "is not available."
            )
            log.warn("         Results might not be biologically relevant!")

        try:
            params = {
                k: v
                for k, v in vars(args).items()
                if k in ["solver", "reference", "multiple_warn_level", "genome"]
            }
            if args.param:
                for pl in args.param:
                    for p in pl:
                        if "=" not in p:
                            raise AldyException(f"Invalid parameter {p}")
                        k, v = p.split("=", 1)
                        params[k.replace("-", "_")] = v
            _ = genotype(
                gene_db=gene,
                sam_path=args.file,
                profile_name=args.profile,
                output_file=output,
                cn_region=cn_region,
                cn_solution=cn_solution,
                report=True,
                is_simple=args.simple,
                debug=debug,
                **{k: v for k, v in params.items() if v is not None},
            )
        except AldyException as ex:
            log.critical(
                f"ERROR: gene= {gene}, profile= {args.profile}, file= {args.file}"
            )
            log.error(str(ex))

    if args.log:
        fh = logbook.FileHandler(
            args.log, mode="w", bubble=True, level="TRACE"  # type: ignore
        )
        fh.formatter = lambda record, _: record.message  # type: ignore
        fh.push_application()
    if args.debug:
        with tempfile.TemporaryDirectory() as tmp:
            prefix = None
            try:
                prefix = f"{tmp}/{os.path.splitext(os.path.basename(args.file))[0]}"
                log_output = f"{prefix}.log"
                fh = logbook.FileHandler(
                    log_output, mode="w", bubble=True, level="TRACE"  # type: ignore
                )
                fh.formatter = lambda record, _: "[{}:{}/{}] {}".format(  # type: ignore
                    record.level_name[0],
                    os.path.splitext(os.path.basename(record.filename))[0],
                    record.func_name,
                    record.message,
                )
                fh.push_application()
                log.debug(f"Using {tmp} as temporary debug directory")
                run(prefix)
            finally:
                if prefix:
                    log.info("Preparing debug archive...")
                    with open(f"{prefix}.yml", "w") as f:
                        yaml.Dumper.ignore_aliases = lambda *_: True  # type: ignore
                        yaml.dump(common.json, f, default_flow_style=None)
                    os.system(f"tar czf {args.debug}.tar.gz -C {tmp} .")
    else:
        run(None)


def _run_test() -> None:
    """
    Run the Aldy test suite.
    """

    pytest.main(["--pyargs", "aldy"])


def console():
    main(sys.argv[1:])


if __name__ == "__main__":
    console()




#cython: boundscheck=False, wraparound=False
from typing import (
    NamedTuple,
    Union
)

from cython.operator cimport postincrement as inc
from libc.stdint cimport int32_t, uint32_t, uint16_t, int8_t, uint8_t

"""
What is a CIGAR?
http://genome.sph.umich.edu/wiki/SAM#What_is_a_CIGAR.3F
"""

cdef int8_t* DNA_BASE_LUT = [
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 0, 4, 1,  4, 4, 4, 2,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  3, 0, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 0, 4, 1,  4, 4, 4, 2,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  3, 0, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
]

cdef inline void dnaToInt8(const char *c_str, int8_t *arr, int len):
    for i in range(0, len):
        arr[i] = DNA_BASE_LUT[c_str[i]]

cdef inline str _str(s):
    if isinstance(s, bytes):
        # encode to the specific encoding used inside of the module
        return (<bytes>s).decode('utf8')
    else:
        return s

cdef extern from "Python.h":
    cdef int PyBytes_Check(object)
    cdef int PyBytes_AsStringAndSize(object, char **, Py_ssize_t *)
    cdef char* PyUnicode_AsUTF8AndSize(object, Py_ssize_t *)
    cdef void* PyMem_Malloc(size_t)
    cdef void PyMem_Free(void*)

cdef inline char* obj_to_cstr_len(object o1, Py_ssize_t *length):
    cdef char* c_str1
    if PyBytes_Check(o1):
        if PyBytes_AsStringAndSize(o1, &c_str1, length) == -1:
            raise TypeError("obj_to_cstr_len: PyBytes_AsStringAndSize error")
        return c_str1
    else:
        c_str1 = PyUnicode_AsUTF8AndSize(o1, length)
        if c_str1 == NULL:
            raise OSError("obj_to_cstr_len: PyUnicode_AsUTF8AndSize error")
    return c_str1

cdef extern from "ssw.h":
    # leave out a few members
    ctypedef struct s_profile:
        const int8_t* read
        const int8_t* mat
        int32_t readLen
        int32_t n
        uint8_t bias

    ctypedef struct s_align:
        uint16_t score1
        uint16_t score2
        int32_t ref_begin1
        int32_t ref_end1
        int32_t read_begin1
        int32_t read_end1
        int32_t ref_end2
        uint32_t* cigar
        int32_t cigarLen

    s_profile* ssw_init(const int8_t*, const int32_t, const int8_t*, const int32_t, const int8_t)
    void init_destroy (s_profile*)
    s_align* ssw_align (const s_profile*, const int8_t*, int32_t, const uint8_t, const uint8_t, const uint8_t, const uint16_t, const int32_t, const int32_t)
    void align_destroy (s_align*)
    char cigar_int_to_op (uint32_t)
    uint32_t cigar_int_to_len(uint32_t)
    uint32_t to_cigar_int(uint32_t, char)

Alignment = NamedTuple("Alignment", [
        ('CIGAR', str),
        ('optimal_score', int),
        ('sub_optimal_score', int),
        ('reference_start', int),
        ('reference_end', int),
        ('read_start', int),
        ('read_end', int)
    ]
)

STR_T = Union[str, bytes]


cdef class SSW:

    cdef int8_t* score_matrix
    cdef s_profile* profile

    cdef object read
    cdef int8_t* read_arr
    cdef Py_ssize_t read_length

    cdef object reference
    cdef int8_t* ref_arr
    cdef Py_ssize_t ref_length

    def __cinit__(self, int match_score=2, int mismatch_penalty=2):
        self.score_matrix = NULL
        self.profile = NULL
        self.read_arr = NULL
        self.ref_arr = NULL
    # end

    def __init__(self,  int match_score=2,
                        int mismatch_penalty=2):
        """ Requires a

        Args:
            match_score (int): for scoring matches
            mismatch_penalty (int): for scoring mismatches
        """
        self.score_matrix = <int8_t*> PyMem_Malloc(25*sizeof(int8_t))
        self.buildDNAScoreMatrix(   <uint8_t>match_score,
                                    <uint8_t> mismatch_penalty,
                                    self.score_matrix)
        self.read = None
        self.reference = None
    # end def

    def __dealloc__(self):
        PyMem_Free(self.score_matrix)

        if self.profile != NULL:
            init_destroy(self.profile)
            self.profile = NULL

        if self.read_arr != NULL:
            PyMem_Free(self.read_arr)

        if self.ref_arr != NULL:
            PyMem_Free(self.ref_arr)
    # end def

    def setRead(self, read: STR_T):
        """ Set the query read string

        Args:
            read:  String-like (str or bytestring) that represents the read.
                    Must be set
        """
        cdef Py_ssize_t read_length
        cdef const char* read_cstr = obj_to_cstr_len(read, &read_length)
        cdef int8_t* read_arr = <int8_t*> PyMem_Malloc(read_length*sizeof(char))

        if self.profile != NULL:
            init_destroy(self.profile)
            self.profile = NULL
        if self.read_arr != NULL:
            PyMem_Free(self.read_arr)
            self.read_arr = NULL

        self.read = read
        self.read_arr = read_arr
        self.read_length = read_length
        dnaToInt8(read_cstr, read_arr, read_length)

        self.profile = ssw_init(read_arr,
                                <int32_t> read_length,
                                self.score_matrix,
                                5,
                                2 # don't know best score size
                                )
    # end def

    def setReference(self, reference: STR_T):
        """Set the query reference string

        Args:
            reference:  String-like (str or bytestring) that represents the
                reference sequence must be set
        """
        cdef Py_ssize_t ref_length
        cdef const char* ref_cstr = obj_to_cstr_len(reference, &ref_length)
        cdef int8_t* ref_arr = <int8_t*> PyMem_Malloc(ref_length*sizeof(char))
        dnaToInt8(ref_cstr, ref_arr, ref_length)
        self.reference = reference
        if self.ref_arr != NULL:
            PyMem_Free(self.ref_arr)
            self.ref_arr = NULL
        self.ref_arr = ref_arr
        self.ref_length = ref_length
    # end def

    cdef s_align* align_c(self,
        int gap_open,
        int gap_extension,
        Py_ssize_t start_idx,
        int32_t mod_ref_length) except NULL:
        """C version of the alignment code
        """
        cdef Py_ssize_t read_length
        cdef const char* read_cstr = obj_to_cstr_len(self.read, &read_length)
        cdef s_align* result = NULL
        cdef int32_t mask_len = read_length // 2

        mask_len = 15 if mask_len < 15 else mask_len

        if self.profile != NULL:
            result = ssw_align ( self.profile,
                                &self.ref_arr[start_idx],
                                mod_ref_length,
                                gap_open,
                                gap_extension,
                                1, 0, 0, mask_len)
        else:
            raise ValueError("Must set profile first")
        if result == NULL:
            raise ValueError("Problem Running alignment, see stdout")
        return result
    # end def

    def align(self,
        int gap_open = 3,
        int gap_extension = 1,
        Py_ssize_t start_idx = 0,
        Py_ssize_t end_idx = 0) -> Alignment:
        '''Align a read to the reference with optional index offseting

        returns a dictionary no matter what as align_c can't return
        NULL

        Args:
            gap_open (int):         penalty for gap_open. default 3
            gap_extension (int):    penalty for gap_extension. default 1
            start_idx (Py_ssize_t): index to start search. default 0
            end_idx (Py_ssize_t):   index to end search (trying to avoid a target region).
                                    default 0 means use whole reference length

        Returns:
            Alignment with keys `CIGAR`,        <for depicting alignment>
                                `optimal_score`,
                                `sub-optimal_score`,
                                `reference_start`, <index into reference>
                                `reference_end`,   <index into reference>
                                `read_start`,  <index into read>
                                `read_end`     <index into read>

        Raises
            ValueError
        '''
        cdef Py_ssize_t c
        cdef char letter
        cdef int letter_int
        cdef uint32_t length
        cdef int32_t search_length
        cdef Py_ssize_t end_idx_final

        if start_idx < 0 or end_idx < 0:
            raise ValueError("negative indexing not supported")
        if end_idx > self.ref_length or start_idx > self.ref_length:
            err = "start_idx: {} or end_idx: {} can't be greater than ref_length: {}".format(
                                                start_idx,
                                                end_idx,
                                                self.ref_length)
            raise ValueError(err)
        if end_idx == 0:
            end_idx_final = self.ref_length
        else:
            end_idx_final = end_idx
        search_length = end_idx_final - start_idx

        cigar = None

        if self.reference is None:
            raise ValueError("call setReference first")

        cdef s_align* result =  self.align_c(gap_open, gap_extension, start_idx, search_length)
        if result.cigar != NULL:
            cigar = ""
            for c in range(result.cigarLen):
                letter = cigar_int_to_op(result.cigar[c])
                letter_int = letter
                length = cigar_int_to_len(result.cigar[c])
                cigar += "%d%s" % (<int>length, chr(letter_int))
        out = Alignment(
                cigar,
                result.score1,
                result.score2,
                result.ref_begin1,
                result.ref_end1,
                result.read_begin1,
                result.read_end1
        )
        #print("RAW BEGIN")
        #self.printResult_c(result)
        #print("RAW END")
        align_destroy(result)
        return out
    # end def

    cdef int buildDNAScoreMatrix(self,
        const uint8_t match_score,
        const uint8_t mismatch_penalty,
        int8_t* matrix) except -1:
        """
        mismatch_penalty should be positive

        The score matrix looks like
                            A,  C,  G,  T,  N
        score_matrix  = {   2, -2, -2, -2,  0, // A
                           -2,  2, -2, -2,  0, // C
                           -2, -2,  2, -2,  0, // G
                           -2, -2, -2,  2,  0, // T
                            0,  0,  0,  0,  0  // N
                        }
        """
        cdef Py_ssize_t i, j
        cdef Py_ssize_t idx = 0;
        for i in range(4):
            for j in range(4):
                if i == j:
                    matrix[idx] =  <int8_t> match_score
                else:
                    matrix[idx] = <int8_t> (-mismatch_penalty)
                inc(idx)
            matrix[idx] = 0;
            inc(idx)
        for i in range(5):
            matrix[inc(idx)] = 0
        return 0
    # end def
# end class

def force_align( read: STR_T,
                reference: STR_T,
                force_overhang: bool = False,
                aligner: SSW = None) -> Alignment:
    '''Enforces no gaps by raising the ``gap_open`` penalty

    Args:
        read:
        reference:
        force_overhang: Make sure only one end overhangs
        aligner: pass an existing :class:`SSW` object
    Raises:
        ValueError for no solution found
    '''
    a: SSW = SSW() if aligner is None else aligner
    a.setRead(read)
    a.setReference(reference)
    len_x: int = len(read)
    # set the gap_open penalty high to drop all splits in alignment
    # pick the first max hit
    res: Alignment = a.align(gap_open=len_x)
    if res.optimal_score < 4:
        raise ValueError("No solution found")
    if force_overhang:
        # read must align to either the beginning or end of the reference string
        if (res.reference_start != 0 or
            res.reference_end != len(reference) - 1):
            raise ValueError("Read does not align to one overhang")
    return res
# end def

def format_force_align(  read: STR_T,
                        reference: STR_T,
                        alignment: Alignment,
                        do_print: bool = False):
    '''Does not truncate strings

    Args:
        read:
        reference:
        alignment:
        do_print: default is False
    '''
    start_ref: int = alignment.reference_start
    start_read: int = alignment.read_start
    buffer_ref: str = ''
    buffer_read: str = ''
    if start_ref < start_read:
        buffer_ref = ' '*(start_read - start_ref)
    else:
        buffer_read = ' '*(start_ref - start_read)
    ref_out = buffer_ref + _str(reference)
    read_out = buffer_read + _str(read)
    if do_print:
        print(ref_out)
        print(read_out)
    return ref_out, read_out
# end def



from pysam.libcfaidx cimport FastaFile


cdef class UnsplicedLocalReference:
    
    def __init__(self, str chrom, int pos, int ref_len, int window, FastaFile reference):
            
        cdef:
            int local_ref_start = max(0, pos - window * 10) # 0-based coordinate
            str unspliced_local_reference = reference.fetch(chrom, local_ref_start, min(pos + window * 10, ref_len))
        
        self.chrom = chrom
        self.pos = pos
        self.ref_len = ref_len
        self.window = window
        self.local_ref_start = local_ref_start
        self.unspliced_local_reference = unspliced_local_reference

    def fetch_ref_seq(self, target_pos, window):
        #pos_diff = target_pos - self.pos
        #lt_end = self.window + pos_diff
        #local_ref_len = min(target_pos + self.window * 3, self.ref_len) - max(0, target_pos - self.window * 3)
        
        self.left_len = target_pos - max(0, target_pos - window * 3)
        return self.get_ref_seq(max(0, target_pos - window * 3), min(target_pos + window * 3, self.ref_len))
        

    def get_ref_seq(self, start, end):
        # start, end: 0-based
        start_idx = start - self.local_ref_start
        return self.unspliced_local_reference[start_idx : start_idx + (end - start)]
        

        
